from datetime import datetime, timedelta, timezone
from typing import List, Union, Sequence, Any, AsyncGenerator, Optional, Type, Mapping

import bson
from beanie.odm.operators.find.comparison import In
from beanie.odm.operators.find.logical import And
from beanie.odm.operators.update.array import AddToSet
from beanie.odm.operators.update.general import Set
from nonebot import logger
from pymongo import UpdateOne

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.enums import RankingMode
from nonebot_plugin_pixivbot.model import Illust, User
from .mongo_models import UserDetailCache, PixivRepoCache, IllustDetailCache, IllustSetCache, UserSetCache, \
    SearchIllustCache, SearchUserCache, UserIllustsCache, UserBookmarksCache, OtherIllustCache, RelatedIllustsCache, \
    IllustRankingCache, DownloadCache
from ..errors import CacheExpiredError, NoSuchItemError
from ..lazy_illust import LazyIllust
from ..models import PixivRepoMetadata
from ...local_tag import LocalTagRepo
from ...source import MongoDataSource


def _handle_expires_in(metadata: PixivRepoMetadata, expires_in: int):
    if datetime.now(timezone.utc) - metadata.update_time >= timedelta(seconds=expires_in):
        raise CacheExpiredError(metadata)


@context.inject
@context.register_singleton()
class MongoPixivRepo:
    conf: Config = Inject(Config)
    mongo: MongoDataSource = Inject(MongoDataSource)
    local_tag_repo: LocalTagRepo = Inject(LocalTagRepo)

    async def _add_to_local_tags(self, illusts: List[Union[LazyIllust, Illust]]):
        li = []
        for x in illusts:
            if isinstance(x, LazyIllust):
                if not x.loaded:
                    continue
                x = x.content
            li.append(x)

        await self.local_tag_repo.update_from_illusts(li)

    async def _illusts_agen(self, doc_type: Type[PixivRepoCache],
                            *criteria: Union[Mapping[str, Any], bool],
                            offset: int = 0) -> AsyncGenerator[LazyIllust, None]:
        aggregation = [
            {
                "$match": And(*criteria).query
            },
            {
                "$replaceWith": {"illust_id": "$illust_id"}
            },
            {
                "$unwind": "$illust_id"
            },
        ]

        if offset:
            aggregation.append({"$offset": offset})

        aggregation.extend([
            {
                "$lookup": {
                    "from": IllustDetailCache.Settings.name,
                    "localField": "illust_id",
                    "foreignField": "illust.id",
                    "as": "illusts"
                }
            },
            {
                "$replaceWith": {
                    "$mergeObjects": [
                        "$$ROOT",
                        {"$arrayElemAt": ["$illusts", 0]}
                    ]
                }
            },
            {
                "$project": {"_id": 0, "illust": 1, "illust_id": 1}
            }
        ])

        total = 0
        broken = 0

        try:
            async for x in doc_type.aggregate(aggregation):
                total += 1
                if "illust" in x and x["illust"] is not None:
                    yield LazyIllust(x["illust_id"], Illust.parse_obj(x["illust"]))
                else:
                    yield LazyIllust(x["illust_id"])
                    broken += 1
        finally:
            logger.debug(f"[local] {total} got, illust_detail of {broken} are missed")

    async def _users_agen(self, doc_type: Type[PixivRepoCache],
                          *criteria: Union[Mapping[str, Any], bool],
                          offset: int = 0) -> AsyncGenerator[User, None]:
        aggregation = [
            {
                "$match": And(*criteria).query
            },
            {
                "$replaceWith": {"user_id": "$user_id"}
            },
            {
                "$unwind": "$user_id"
            },
        ]

        if offset:
            aggregation.append({"$offset": offset})

        aggregation.extend([
            {
                "$lookup": {
                    "from": UserDetailCache.Settings.name,
                    "localField": "user_id",
                    "foreignField": "user.id",
                    "as": "users"
                }
            },
            {
                "$replaceWith": {
                    "$mergeObjects": [
                        "$$ROOT",
                        {"$arrayElemAt": ["$users", 0]}
                    ]
                }
            },
            {
                "$project": {"_id": 0, "user": 1, "user_id": 1}
            }
        ])

        async for x in doc_type.aggregate(aggregation):
            if "user" in x and x["user"] is not None:
                yield User.parse_obj(x["user"])
            else:
                yield User(id=x["user_id"], name="", account="")

    async def _get_illusts(self, doc_type: Type[PixivRepoCache],
                           *criteria: Union[Mapping[str, Any], bool],
                           expired_in: int,
                           offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], Any]:
        doc: Optional[PixivRepoCache] = await doc_type.find_one(*criteria)
        if not doc:
            raise NoSuchItemError()
        _handle_expires_in(doc.metadata, expired_in)

        metadata = doc.metadata
        yield metadata.copy(update={"pages": 0})
        async for x in self._illusts_agen(doc_type, *criteria, offset=offset):
            yield x
        yield metadata

    async def _get_users(self, doc_type: Type[PixivRepoCache],
                         *criteria: Union[Mapping[str, Any], bool],
                         expired_in: int,
                         offset: int = 0) \
            -> AsyncGenerator[Union[User, PixivRepoMetadata], Any]:
        doc: Optional[PixivRepoCache] = await doc_type.find_one(*criteria)
        if not doc:
            raise NoSuchItemError()
        _handle_expires_in(doc.metadata, expired_in)

        metadata = doc.metadata
        yield metadata.copy(update={"pages": 0})
        async for x in self._users_agen(doc_type, *criteria, offset=offset):
            yield x
        yield metadata

    async def _check_illusts_exists(self, doc_type: Type[IllustSetCache],
                                    *criteria: Union[Mapping[str, Any], bool],
                                    illust_id: Union[int, Sequence[int]]) -> bool:
        if isinstance(illust_id, int):
            return await doc_type.find(*criteria, In(doc_type.illust_id, illust_id)).count() != 0
        else:
            agg = [
                {'$match': And(*criteria).query},
                {'$unwind': {'path': '$illust_id'}},
                {'$match': {'illust_id': {'$in': illust_id}}},
                {'$count': 'count'}
            ]

            async for result in doc_type.aggregate(agg):
                return result["count"] != 0

    async def _check_users_exists(self, doc_type: Type[UserSetCache],
                                  *criteria: Union[Mapping[str, Any], bool],
                                  user_id: Union[int, Sequence[int]]) -> bool:
        if isinstance(user_id, int):
            return await doc_type.find(*criteria, In(doc_type.user_id, user_id)).count() != 0
        else:
            agg = [
                {'$match': And(*criteria).query},
                {'$unwind': {'path': '$user_id'}},
                {'$match': {'user_id': {'$in': user_id}}},
                {'$count': 'count'}
            ]

            async for result in doc_type.aggregate(agg):
                return result["count"] != 0

    async def _update_illusts(self, doc_type: Type[IllustSetCache],
                              *criteria: Union[Mapping[str, Any], bool],
                              content: List[Union[Illust, LazyIllust]],
                              metadata: PixivRepoMetadata,
                              append: bool = False):
        if append:
            await doc_type.find_one(*criteria).update(
                Set({
                    doc_type.metadata: metadata
                }),
                AddToSet({
                    doc_type.illust_id: {
                        "$each": [illust.id for illust in content]
                    }
                }),
                upsert=True
            )
        else:
            await doc_type.find_one(*criteria).update(
                Set({
                    doc_type.illust_id: [illust.id for illust in content],
                    doc_type.metadata: metadata
                }),
                upsert=True
            )

        # BulkWriter存在bug，upsert不生效
        # https://github.com/roman-right/beanie/issues/224
        #
        # async with BulkWriter() as bw:
        #     illust_metadata = PixivRepoMetadata(update_time=metadata.update_time)
        #
        #     for illust in content:
        #         if isinstance(illust, LazyIllust) and illust.content is not None:
        #             illust = illust.content
        #
        #         if isinstance(illust, Illust):
        #             await IllustDetailCache.find_one(
        #                 IllustDetailCache.illust.id == illust.id
        #             ).upsert(
        #                 Set({
        #                     IllustDetailCache.illust: illust,
        #                     IllustDetailCache.metadata: illust_metadata
        #                 }),
        #                 on_insert=IllustDetailCache(illust=illust, metadata=illust_metadata),
        #                 bulk_writer=bw
        #             )

        opt = []
        for illust in content:
            if isinstance(illust, LazyIllust) and illust.content is not None:
                illust = illust.content

            if isinstance(illust, Illust):
                opt.append(UpdateOne(
                    {"illust.id": illust.id},
                    {"$set": {
                        "illust": illust.dict(exclude_none=True),
                        "metadata": {"update_time": metadata.update_time}
                    }},
                    upsert=True
                ))
        if len(opt) != 0:
            await self.mongo.db.illust_detail_cache.bulk_write(opt, ordered=False)

        if self.conf.pixiv_tag_translation_enabled:
            await self._add_to_local_tags(content)

    async def _update_users(self, doc_type: Type[UserSetCache],
                            *criteria: Union[Mapping[str, Any], bool],
                            content: List[User],
                            metadata: PixivRepoMetadata,
                            append: bool = False):
        if append:
            await doc_type.find_one(*criteria).update(
                Set({
                    doc_type.metadata: metadata
                }),
                AddToSet({
                    doc_type.user_id: {
                        "$each": [user.id for user in content]
                    }
                }),
                upsert=True
            )
        else:
            await doc_type.find_one(*criteria).update(
                Set({
                    doc_type.user_id: [user.id for user in content],
                    doc_type.metadata: metadata
                }),
                upsert=True
            )
        # BulkWriter存在bug，upsert不生效
        # https://github.com/roman-right/beanie/issues/224
        #
        # async with BulkWriter() as bw:
        #     user_metadata = PixivRepoMetadata(update_time=metadata.update_time)
        #
        #     for user in content:
        #         await UserDetailCache.find_one(
        #             UserDetailCache.user.id == user.id
        #         ).upsert(
        #             Set({
        #                 UserDetailCache.user: user,
        #                 UserDetailCache.metadata: user_metadata
        #             }),
        #             on_insert=UserDetailCache(user=user, metadata=user_metadata),
        #             bulk_writer=bw
        #         )

        opt = []
        for user in content:
            opt.append(UpdateOne(
                {"user.id": user.id},
                {"$set": {
                    "user": user.dict(exclude_none=True),
                    "metadata": {"update_time": metadata.update_time}
                }},
                upsert=True
            ))
        if len(opt) != 0:
            await self.mongo.db.user_detail_cache.bulk_write(opt, ordered=False)

    async def _append_and_check_illusts(self, doc_type: Type[IllustSetCache],
                                        *criteria: Union[Mapping[str, Any], bool],
                                        content: List[Union[Illust, LazyIllust]],
                                        metadata: PixivRepoMetadata):
        exists = await self._check_illusts_exists(doc_type, *criteria, illust_id=[x.id for x in content])
        await self._update_illusts(doc_type, *criteria, content=content, metadata=metadata, append=True)
        return exists

    async def _append_and_check_users(self, doc_type: Type[UserSetCache],
                                      *criteria: Union[Mapping[str, Any], bool],
                                      content: List[User],
                                      metadata: PixivRepoMetadata):
        exists = await self._check_users_exists(doc_type, *criteria, user_id=[x.id for x in content])
        await self._update_users(doc_type, *criteria, content=content, metadata=metadata, append=True)
        return exists

    # ================ illust_detail ================
    async def illust_detail(self, illust_id: int) \
            -> AsyncGenerator[Union[Illust, PixivRepoMetadata], None]:
        logger.debug(f"[local] illust_detail {illust_id}")
        doc = await IllustDetailCache.find_one(IllustDetailCache.illust.id == illust_id)
        if doc is not None:
            _handle_expires_in(doc.metadata, self.conf.pixiv_illust_detail_cache_expires_in)

            yield doc.metadata
            yield doc.illust
        else:
            raise NoSuchItemError()

    async def update_illust_detail(self, illust: Illust, metadata: PixivRepoMetadata):
        logger.debug(f"[local] update illust_detail {illust.id} {metadata}")

        await IllustDetailCache.find_one(
            IllustDetailCache.illust.id == illust.id
        ).update(
            Set({
                IllustDetailCache.illust: illust,
                IllustDetailCache.metadata: metadata
            }),
            upsert=True
        )

        if self.conf.pixiv_tag_translation_enabled:
            await self._add_to_local_tags([illust])

    # ================ user_detail ================
    async def user_detail(self, user_id: int) \
            -> AsyncGenerator[Union[User, PixivRepoMetadata], None]:
        logger.debug(f"[local] user_detail {user_id}")
        doc: Optional[UserDetailCache] = await UserDetailCache.find_one(UserDetailCache.user.id == user_id)
        if doc is not None:
            _handle_expires_in(doc.metadata, self.conf.pixiv_user_detail_cache_expires_in)

            yield doc.metadata
            yield doc.user
        else:
            raise NoSuchItemError()

    async def update_user_detail(self, user: User, metadata: PixivRepoMetadata):
        logger.debug(f"[local] update user_detail {user.id} {metadata}")

        if not metadata:
            metadata = PixivRepoMetadata(update_time=datetime.now(timezone.utc))

        await UserDetailCache.find_one(
            UserDetailCache.user.id == user.id
        ).update(
            Set({
                UserDetailCache.user: user,
                UserDetailCache.metadata: metadata
            }),
            upsert=True
        )

    # ================ search_illust ================
    def search_illust(self, word: str, *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[local] search_illust {word}")
        return self._get_illusts(SearchIllustCache, SearchIllustCache.word == word,
                                 expired_in=self.conf.pixiv_search_illust_cache_expires_in, offset=offset)

    async def invalidate_search_illust(self, word: str):
        logger.debug(f"[local] invalidate search_illust {word}")
        await self.mongo.db.search_illust_cache.delete_one({"word": word})

    async def append_search_illust(self, word: str, content: List[Union[Illust, LazyIllust]],
                                   metadata: PixivRepoMetadata) -> bool:
        # 返回值表示content中是否有已经存在于集合的文档，下同
        logger.debug(f"[local] append search_illust {word} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts(SearchIllustCache, SearchIllustCache.word == word,
                                                    content=content, metadata=metadata)

    # ================ search_user ================

    def search_user(self, word: str, *, offset: int = 0) \
            -> AsyncGenerator[Union[User, PixivRepoMetadata], None]:
        logger.debug(f"[local] search_user {word}")
        return self._get_users(SearchUserCache, SearchUserCache.word == word,
                               expired_in=self.conf.pixiv_search_user_cache_expires_in, offset=offset)

    async def invalidate_search_user(self, word: str):
        logger.debug(f"[local] invalidate search_user {word}")
        await SearchUserCache.find_one(SearchUserCache.word == word).delete()

    async def append_search_user(self, word: str, content: List[User],
                                 metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append search_user {word} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_users(SearchUserCache, SearchUserCache.word == word,
                                                  content=content, metadata=metadata)

    # ================ user_illusts ================

    def user_illusts(self, user_id: int, *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[local] user_illusts {user_id}")
        return self._get_illusts(UserIllustsCache, UserIllustsCache.user_id == user_id,
                                 expired_in=self.conf.pixiv_user_illusts_cache_expires_in, offset=offset)

    async def invalidate_user_illusts(self, user_id: int):
        logger.debug(f"[local] invalidate user_illusts {user_id}")
        await UserIllustsCache.find_one(UserIllustsCache.user_id == user_id).delete()

    async def append_user_illusts(self, user_id: int,
                                  content: List[Union[Illust, LazyIllust]],
                                  metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append user_illusts {user_id} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts(UserIllustsCache, UserIllustsCache.user_id == user_id,
                                                    content=content, metadata=metadata)

    # ================ user_bookmarks ================

    def user_bookmarks(self, user_id: int = 0, *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[local] user_bookmarks {user_id}")
        return self._get_illusts(UserBookmarksCache, UserBookmarksCache.user_id == user_id,
                                 expired_in=self.conf.pixiv_user_bookmarks_cache_expires_in, offset=offset)

    async def invalidate_user_bookmarks(self, user_id: int):
        logger.debug(f"[local] invalidate user_bookmarks {user_id}")
        await UserBookmarksCache.find_one(UserBookmarksCache.user_id == user_id).delete()

    async def append_user_bookmarks(self, user_id: int,
                                    content: List[Union[Illust, LazyIllust]],
                                    metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append user_bookmarks {user_id} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts(UserBookmarksCache, UserBookmarksCache.user_id == user_id,
                                                    content=content, metadata=metadata)

    # ================ recommended_illusts ================

    def recommended_illusts(self, *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[local] recommended_illusts")
        return self._get_illusts(OtherIllustCache, OtherIllustCache.type == "recommended_illusts",
                                 expired_in=self.conf.pixiv_other_cache_expires_in, offset=offset)

    async def invalidate_recommended_illusts(self):
        logger.debug(f"[local] invalidate recommended_illusts")
        await OtherIllustCache.find_one(OtherIllustCache.type == "recommended_illusts").delete()

    async def append_recommended_illusts(self, content: List[Union[Illust, LazyIllust]],
                                         metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append recommended_illusts "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts(OtherIllustCache, OtherIllustCache.type == "recommended_illusts",
                                                    content=content, metadata=metadata)

    # ================ related_illusts ================

    def related_illusts(self, illust_id: int, *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[local] related_illusts {illust_id}")
        return self._get_illusts(RelatedIllustsCache, RelatedIllustsCache.original_illust_id == illust_id,
                                 expired_in=self.conf.pixiv_related_illusts_cache_expires_in, offset=offset)

    async def invalidate_related_illusts(self, illust_id: int):
        logger.debug(f"[local] invalidate related_illusts")
        await RelatedIllustsCache.find_one(RelatedIllustsCache.original_illust_id == illust_id).delete()

    async def append_related_illusts(self, illust_id: int, content: List[Union[Illust, LazyIllust]],
                                     metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append related_illusts {illust_id} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts(RelatedIllustsCache,
                                                    RelatedIllustsCache.original_illust_id == illust_id,
                                                    content=content, metadata=metadata)

    # ================ illust_ranking ================
    def illust_ranking(self, mode: Union[str, RankingMode], *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        if isinstance(mode, str):
            mode = RankingMode[mode]

        logger.debug(f"[local] illust_ranking {mode}")

        return self._get_illusts(IllustRankingCache, IllustRankingCache.mode == mode,
                                 expired_in=self.conf.pixiv_illust_ranking_cache_expires_in, offset=offset)

    async def invalidate_illust_ranking(self, mode: RankingMode):
        logger.debug(f"[local] invalidate illust_ranking")
        await IllustRankingCache.find_one(IllustRankingCache.mode == mode).delete()

    async def append_illust_ranking(self, mode: RankingMode, content: List[Union[Illust, LazyIllust]],
                                    metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append illust_ranking {mode} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts(IllustRankingCache, IllustRankingCache.mode == mode,
                                                    content=content, metadata=metadata)

    # ================ image ================

    async def image(self, illust: Illust) -> AsyncGenerator[Union[bytes, PixivRepoMetadata], None]:
        logger.debug(f"[local] image {illust.id}")
        doc = await DownloadCache.find_one(DownloadCache.illust_id == illust.id)
        if doc is not None:
            _handle_expires_in(doc.metadata, self.conf.pixiv_download_cache_expires_in)
            yield doc.metadata
            yield doc.content
        else:
            raise NoSuchItemError()

    async def update_image(self, illust_id: int, content: bytes,
                           metadata: PixivRepoMetadata):
        logger.debug(f"[local] update image {illust_id} "
                     f"{metadata}")

        await DownloadCache.find_one(
            DownloadCache.illust_id == illust_id
        ).update(
            Set({
                DownloadCache.content: bson.Binary(content),
                DownloadCache.metadata: metadata
            }),
            upsert=True
        )

    async def invalidate_all(self):
        logger.debug(f"[local] invalidate_all")
        await self.mongo.db.download_cache.delete_many({})
        await self.mongo.db.illust_detail_cache.delete_many({})
        await self.mongo.db.user_detail_cache.delete_many({})
        await self.mongo.db.illust_ranking_cache.delete_many({})
        await self.mongo.db.search_illust_cache.delete_many({})
        await self.mongo.db.search_user_cache.delete_many({})
        await self.mongo.db.user_illusts_cache.delete_many({})
        await self.mongo.db.user_bookmarks_cache.delete_many({})
        await self.mongo.db.other_cache.delete_many({})


__all__ = ("MongoPixivRepo",)



import asyncio
from typing import Callable, TypeVar

from nonebot import logger
from nonebot.exception import ActionFailed

from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.utils.errors import BadRequestError, QueryError
from .base import Interceptor
from ..pkg_context import context

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.register_singleton()
class DefaultErrorInterceptor(Interceptor):
    async def intercept(self, wrapped_func: Callable, *args,
                        post_dest: PostDestination[UID, GID],
                        silently: bool,
                        **kwargs):
        try:
            await wrapped_func(*args, post_dest=post_dest, silently=silently, **kwargs)
        except asyncio.TimeoutError:
            logger.warning("Timeout")
            if not silently:
                await self.post_plain_text(f"下载超时", post_dest=post_dest)
        except BadRequestError as e:
            if not silently:
                await self.post_plain_text(str(e), post_dest=post_dest)
        except QueryError as e:
            if not silently:
                await self.post_plain_text(str(e), post_dest=post_dest)
        except ActionFailed as e:
            # 避免当发送消息错误时再尝试发送
            raise e
        except Exception as e:
            if not silently:
                await self.post_plain_text(f"内部错误：{type(e)}{e}", post_dest=post_dest)
            raise e  # 重新抛出，让上层可以处理（如scheduler中需要处理Handler的异常）



from typing import Generic, TypeVar

from pydantic.generics import GenericModel

UID = TypeVar("UID")


class PixivBinding(GenericModel, Generic[UID]):
    adapter: str
    user_id: UID
    pixiv_user_id: int



from asyncio import create_task, sleep, shield, Task
from typing import Callable, Optional

from nonebot import logger

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from .base import Interceptor, UID, GID
from ..pkg_context import context


@context.inject
@context.register_singleton()
class LoadingPromptInterceptor(Interceptor):
    conf = Inject(Config)

    async def send_delayed_loading_prompt(self, post_dest: PostDestination[UID, GID]):
        await sleep(self.conf.pixiv_loading_prompt_delayed_time)

        logger.debug(f"send delayed loading to {post_dest.identifier}")
        await shield(self.post_plain_text("努力加载中", post_dest))

    async def intercept(self, wrapped_func: Callable,
                        *args, post_dest: PostDestination[UID, GID],
                        silently: bool,
                        **kwargs):
        task: Optional[Task] = None
        if not silently:
            task = create_task(self.send_delayed_loading_prompt(post_dest))

        try:
            await wrapped_func(*args, post_dest=post_dest, silently=silently, **kwargs)
        finally:
            if task and not task.done():
                task.cancel()



from datetime import datetime, timedelta, timezone

from nonebot import logger

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.pixiv_repo import PixivRepo
from nonebot_plugin_pixivbot.data.pixiv_repo.enums import CacheStrategy
from nonebot_plugin_pixivbot.data.pixiv_repo.remote_repo import RemotePixivRepo
from nonebot_plugin_pixivbot.model import WatchTask, Illust
from nonebot_plugin_pixivbot.utils.shared_agen import SharedAsyncGeneratorManager
from .base import WatchTaskHandler
from ..handler import PD
from ..pkg_context import context


# 因为要强制从远端获取，所以用这个shared_agen_mgr来缓存
@context.inject
@context.register_singleton()
class WatchUserIllustsSharedAsyncGeneratorManager(SharedAsyncGeneratorManager[int, Illust]):
    log_tag = "watch_user_illusts_shared_agen"

    pixiv = Inject(PixivRepo)
    remote_pixiv = Inject(RemotePixivRepo)

    async def agen(self, identifier: int, cache_strategy: CacheStrategy, **kwargs):
        self.set_expires_time(identifier, datetime.now(timezone.utc) + timedelta(seconds=30))  # 保证每分钟的所有task都能共享
        async for x in self.pixiv.user_illusts(user_id=identifier,
                                               cache_strategy=CacheStrategy.FORCE_EXPIRATION):
            yield await x.get()


@context.inject
@context.root.register_singleton()
class WatchUserIllustsHandler(WatchTaskHandler):
    shared_agen_mgr = Inject(WatchUserIllustsSharedAsyncGeneratorManager)

    @classmethod
    def type(cls) -> str:
        return "watch_user_illusts"

    def enabled(self) -> bool:
        return True

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, task: WatchTask, post_dest: PD, silently: bool = False, **kwargs):
        with self.shared_agen_mgr.get(task.kwargs["user_id"]) as illusts:
            async for illust in illusts:
                if illust.create_date <= task.checkpoint:
                    break
                logger.info(f"[watchman] send illust {illust.id} to {task.subscriber}")
                await self.post_illust(illust, header="您订阅的画师更新了", post_dest=post_dest)



from typing import Optional

from nonebot import Bot, get_bot


def get_adapter_name(bot: Optional[Bot] = None) -> str:
    if not bot:
        bot = get_bot()
    return bot.adapter.get_name().split(maxsplit=1)[0].lower()



from typing import TypeVar

from lazy import lazy
from nonebot import on_regex
from nonebot.adapters import Bot, Event
from nonebot.internal.params import Depends
from nonebot.matcher import Matcher
from nonebot.typing import T_State

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from .base import CommonHandler
from ..entry_handler import post_destination
from ..pkg_context import context
from ..recorder import Recorder
from ..utils import get_common_query_rule

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.inject
@context.root.register_eager_singleton()
class MoreHandler(CommonHandler):
    recorder = Inject(Recorder)

    @classmethod
    def type(cls) -> str:
        return "more"

    def enabled(self) -> bool:
        return self.conf.pixiv_more_enabled

    @lazy
    def matcher(self):
        return on_regex("^还要((.*)张)?$", rule=get_common_query_rule(), priority=1, block=True)

    async def on_match(self, bot: Bot, event: Event, state: T_State, matcher: Matcher,
                       post_dest: PostDestination[UID, GID] = Depends(post_destination)):
        illust_id = state["_matched_groups"][0]
        await self.handle(illust_id, post_dest=post_dest)

    async def actual_handle(self, *, count: int = 1,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False,
                            **kwargs):
        req = self.recorder.get_req(post_dest.identifier)
        if not req:
            raise BadRequestError("你还没有发送过请求")

        await req(count=count, post_dest=post_dest, silently=silently)



from typing import TypeVar, Optional, AsyncIterable, Collection

from sqlalchemy import Column, Integer, Enum as SqlEnum, JSON, String, select, DateTime, UniqueConstraint, update, Index
from sqlalchemy.dialects.sqlite import insert

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.seq import SeqRepo
from nonebot_plugin_pixivbot.data.source.sql import SqlDataSource
from nonebot_plugin_pixivbot.data.utils.process_subscriber import process_subscriber
from nonebot_plugin_pixivbot.model import PostIdentifier, WatchType, WatchTask

UID = TypeVar("UID")
GID = TypeVar("GID")

ID = PostIdentifier[UID, GID]


@context.require(SqlDataSource).registry.mapped
class WatchTaskOrm:
    __tablename__ = "watch_task"

    id = Column(Integer, primary_key=True, autoincrement=True)
    subscriber = Column(JSON, nullable=False)
    code = Column(Integer, nullable=False)
    type = Column(SqlEnum(WatchType), nullable=False)
    kwargs = Column(JSON, nullable=False, default=dict)
    adapter = Column(String, nullable=False)
    checkpoint = Column(DateTime, nullable=False)

    __table_args__ = (
        Index("adapter"),
        Index("subscriber", "code"),
        UniqueConstraint("subscriber", "type", "kwargs", sqlite_on_conflict='IGNORE'),
    )


@context.inject
@context.register_singleton()
class SqlWatchTaskRepo:
    data_source: SqlDataSource = Inject(SqlDataSource)
    seq_repo: SeqRepo = Inject(SeqRepo)

    async def get_by_subscriber(self, subscriber: ID) -> AsyncIterable[WatchTask]:
        subscriber = process_subscriber(subscriber)

        session = self.data_source.session()
        stmt = (
            select(WatchTaskOrm)
            .where(WatchTaskOrm.subscriber == subscriber.dict())
        )
        async for x in await session.stream_scalars(stmt):
            yield WatchTask.from_orm(x)

    async def get_by_adapter(self, adapter: str) -> AsyncIterable[WatchTask]:
        session = self.data_source.session()
        stmt = (
            select(WatchTaskOrm)
            .where(WatchTaskOrm.adapter == adapter)
        )
        async for x in await session.stream_scalars(stmt):
            yield WatchTask.from_orm(x)

    async def get_by_code(self, subscriber: ID, code: int) -> Optional[WatchTask]:
        subscriber = process_subscriber(subscriber)

        session = self.data_source.session()
        stmt = (select(WatchTaskOrm)
                .where(WatchTask.subscriber == subscriber.dict(),
                       WatchTask.code == code))
        result = (await session.execute(stmt)).scalar_one_or_none()
        return result

    async def insert(self, task: WatchTask) -> bool:
        task.subscriber = process_subscriber(task.subscriber)

        session = self.data_source.session()
        stmt = (insert(WatchTaskOrm)
                .values(subscriber=task.subscriber.dict(),
                        code=task.code,
                        type=task.type,
                        kwargs=task.kwargs,
                        adapter=task.subscriber.adapter,
                        checkpoint=task.checkpoint))
        result = await session.execute(stmt)
        await session.commit()

        if result.rowcount == 1:
            task.code = await self.seq_repo.inc_and_get(f'watch_task {task.subscriber}')

            task_orm = await session.get(WatchTaskOrm, result.inserted_primary_key[0])
            task_orm.code = task.code
            await session.commit()

            return True
        else:
            return False

    async def update(self, task: WatchTask) -> bool:
        task.subscriber = process_subscriber(task.subscriber)

        session = self.data_source.session()
        stmt = (update(WatchTaskOrm)
                .values(type=task.type,
                        kwargs=task.kwargs,
                        adapter=task.subscriber.adapter,
                        checkpoint=task.checkpoint)
                .where(WatchTask.subscriber == task.subscriber.dict(),
                       WatchTask.code == task.code))
        result = await session.execute(stmt)
        await session.commit()
        return result.rowcount == 1

    async def delete_one(self, subscriber: ID, code: int) -> Optional[WatchTask]:
        subscriber = process_subscriber(subscriber)

        session = self.data_source.session()
        stmt = (select(WatchTaskOrm)
                .where(WatchTaskOrm.subscriber == subscriber.dict(),
                       WatchTaskOrm.code == code)
                .limit(1))
        task = (await session.execute(stmt)).scalar_one_or_none()

        if task is None:
            return task

        await session.delete(task)
        await session.commit()
        return WatchTask.from_orm(task)

    async def delete_many_by_subscriber(self, subscriber: ID) -> Collection[WatchTask]:
        subscriber = process_subscriber(subscriber)

        session = self.data_source.session()
        stmt = (select(WatchTaskOrm)
                .where(WatchTaskOrm.subscriber == subscriber.dict()))
        tasks = (await session.execute(stmt)).scalars().all()

        for t in tasks:
            await session.delete(t)

        await session.commit()
        return tasks


__all__ = ("SqlWatchTaskRepo",)



from abc import ABC, abstractmethod
from typing import TypeVar, Type, Callable, Union, Generic

from nonebot.log import logger

T = TypeVar("T")
T2 = TypeVar("T2")


class Inject:
    def __init__(self, key):
        self._key = key

    def __get__(self, instance, owner):
        context = getattr(instance, "__context__")
        return context.require(self._key)


class Provider(ABC, Generic[T]):
    @abstractmethod
    def provide(self) -> T:
        raise NotImplementedError()


class InstanceProvider(Provider[T], Generic[T]):
    def __init__(self, instance: T):
        self._instance = instance

    def provide(self) -> T:
        return self._instance


class DynamicProvider(Provider[T], Generic[T]):
    def __init__(self, func: Callable[[], T], use_cache: bool = True):
        self._func = func
        self._use_cache = use_cache

        self._cache = None
        self._cached = False

    def provide(self) -> T:
        if not self._use_cache:
            return self._func()

        if not self._cached:
            self._cache = self._func()  # just let it throw
            self._cached = True
        return self._cache


class Context:
    def __init__(self, parent: "Context" = None):
        self._parent = parent
        self._container = {}

    @property
    def parent(self) -> "Context":
        return self._parent

    @property
    def root(self) -> "Context":
        if self._parent is None:
            return self
        else:
            return self._parent.root

    def register(self, key: Type[T], bean: Union[T, Provider[T]]):
        """
        register a bean
        """
        if not isinstance(bean, Provider):
            bean = InstanceProvider(bean)

        self._container[key] = bean
        logger.trace(f"registered bean {key}, provider type: {type(bean)}")

    def register_lazy(self, key: Type[T], bean_initializer: Callable[[], T]):
        """
        register a bean lazily
        """
        self._container[key] = DynamicProvider(bean_initializer)
        logger.trace(f"lazily registered bean {key}")

    def register_singleton(self, *args, **kwargs) -> Callable[[Type[T]], Type[T]]:
        """
        decorator for a class to register a bean lazily
        """

        def decorator(cls: Type[T]) -> Type[T]:
            self.register_lazy(cls, lambda: cls(*args, **kwargs))
            return cls

        return decorator

    def register_eager_singleton(self, *args, **kwargs) -> Callable[[Type[T]], Type[T]]:
        """
        decorator for a class to register a bean lazily
        """

        def decorator(cls: Type[T]) -> Type[T]:
            bean = cls(*args, **kwargs)
            self.register(cls, bean)
            return cls

        return decorator

    def unregister(self, key: Type[T]) -> bool:
        """
        unregister the bean of key
        """
        if key in self._container:
            del self._container[key]
            return True
        return False

    def bind(self, key: Type[T], src_key: Type[T2]):
        """
        bind key (usually the implementation class) to src_key (usually the base class)
        """
        self._container[key] = self._find_provider(src_key)
        logger.trace(f"bind bean {key} to {src_key}")

    def bind_singleton_to(self, key: Type[T], *args, **kwargs) -> Callable[[Type[T2]], Type[T2]]:
        """
        decorator for a class (usually the implementation class) to bind to another class (usually the base class)
        """

        def decorator(cls: Type[T2]) -> Type[T2]:
            self.register_singleton(*args, **kwargs)(cls)
            self.bind(key, cls)
            return cls

        return decorator

    def require(self, key: Type[T]) -> T:
        return self._find_provider(key).provide()

    def _find_provider(self, key: Type[T]) -> T:
        if key in self._container:
            return self._container[key]
        elif self._parent is not None:
            return self._parent._find_provider(key)
        else:
            raise KeyError(key)

    def __getitem__(self, key: Type[T]):
        return self.require(key)

    def __contains__(self, key: Type[T]) -> bool:
        if key in self._container:
            return True
        elif self._parent is not None:
            return self._parent.__contains__(key)
        else:
            return False

    def inject(self, cls: Type[T]):
        setattr(cls, "__context__", self)
        return cls


__all__ = ("Context", "Inject")



from typing import Type

from nonebot import Bot, on_regex
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import Matcher
from nonebot.internal.params import Depends
from nonebot.typing import T_State

from nonebot_plugin_pixivbot.handler.common import IllustHandler
from nonebot_plugin_pixivbot.handler.entry_handler import DelegationEntryHandler, UID, GID, EntryHandler, \
    post_destination
from nonebot_plugin_pixivbot.handler.utils import get_common_query_rule
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from ..pkg_context import context


@context.root.register_eager_singleton()
class IllustLinkHandler(DelegationEntryHandler):
    @classmethod
    def type(cls) -> str:
        return "illust_link"

    def enabled(self) -> bool:
        return True

    def delegation(self) -> EntryHandler:
        return context.require(IllustHandler)

    @property
    def matcher(self) -> Type[Matcher]:
        return on_regex(r"^(http://|https://)?(www.)?pixiv\.net/artworks/([1-9][0-9]*)/?$",
                        rule=get_common_query_rule(), priority=5)

    async def on_match(self, bot: Bot, event: Event, state: T_State, matcher: Matcher,
                       post_dest: PostDestination[UID, GID] = Depends(post_destination)):
        illust_id = state["_matched_groups"][2]
        await self.handle(illust_id, post_dest=post_dest)



from __future__ import annotations

import time
from collections import OrderedDict
from typing import Optional, TypeVar

from nonebot import logger

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.model import PostIdentifier
from nonebot_plugin_pixivbot.model.message import IllustMessageModel, IllustMessagesModel
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.protocol_dep.postman import PostmanManager
from .pkg_context import context

UID = TypeVar("UID")
GID = TypeVar("GID")

PD = PostDestination[UID, GID]
ID = PostIdentifier[UID, GID]


class Req:
    def __init__(self, func,
                 *args, **kwargs):
        self.timestamp = 0
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.refresh()

    def refresh(self):
        self.timestamp = time.time()

    def __call__(self, **kwargs):
        actual_kwargs = self.kwargs.copy()
        for x, y in kwargs.items():
            actual_kwargs[x] = y

        return self.func(*self.args, **actual_kwargs)


class Resp:
    def __init__(self, illust_id: int):
        self.timestamp = 0
        self.illust_id = illust_id
        self.refresh()

    def refresh(self):
        self.timestamp = time.time()


@context.inject
@context.root.register_singleton()
class Recorder:
    conf = Inject(Config)

    def __init__(self, max_req_size: int = 65535,
                 max_resp_size: int = 65535):
        self._reqs = OrderedDict[ID, Req]()
        self._resps = OrderedDict[ID, Resp]()

        self.max_req_size = max_req_size
        self.max_resp_size = max_resp_size

    @staticmethod
    def _key_fallback(key: ID) -> Optional[ID]:
        if key.user_id is not None and key.group_id is not None:
            return ID(key.adapter, None, key.group_id)
        else:
            return None

    def _collate_req(self, ensure_size_for_put: bool):
        now = time.time()
        while len(self._reqs) > 0:
            key, req = next(iter(self._reqs.items()))
            if now - req.timestamp > self.conf.pixiv_query_expires_in:
                self._reqs.popitem(last=False)
                logger.info(f"popped expired req: ({key})")
            else:
                break

        if ensure_size_for_put and len(self._reqs) == self.max_req_size:
            key, req = self._reqs.popitem(last=False)
            logger.info(f"popped first req: ({key})")

    def record_req(self, record: Req, key: ID):
        self._collate_req(True)
        if key in self._reqs:
            self._reqs.move_to_end(key)
        self._reqs[key] = record

    def get_req(self, key: ID) -> Optional[Req]:
        self._collate_req(False)
        if key in self._reqs:
            req = self._reqs[key]
            req.refresh()
            self._reqs.move_to_end(key)
            return req
        else:
            key = self._key_fallback(key)
            if key is not None:
                return self.get_req(key)
            else:
                return None

    def _collate_resp(self, ensure_size_for_put: bool):
        now = time.time()
        while len(self._resps) > 0:
            key, resp = next(iter(self._resps.items()))
            if now - resp.timestamp > self.conf.pixiv_query_expires_in:
                self._resps.popitem(last=False)
                logger.info(f"popped expired resp illust: ({key})")
            else:
                break

        if ensure_size_for_put and len(self._resps) == self.max_resp_size:
            key, resp = self._resps.popitem(last=False)
            logger.info(f"popped first resp illust: ({key})")

    def record_resp(self, illust_id: int, key: ID):
        self._collate_resp(True)
        if key in self._resps:
            self._resps.move_to_end(key)
        self._resps[key] = Resp(illust_id)

    def get_resp(self, key: ID) -> Optional[int]:
        self._collate_resp(False)
        if key in self._resps:
            rec = self._resps[key]
            return rec.illust_id
        else:
            key = self._key_fallback(key)
            if key is not None:
                return self.get_resp(key)
            else:
                return None


@context.inject
@context.bind_singleton_to(PostmanManager, context.require(PostmanManager))
class RecordPostmanManager:
    recorder = Inject(Recorder)

    def __init__(self, delegation: PostmanManager):
        self.delegation = delegation

    async def send_illust(self, model: IllustMessageModel,
                          *, post_dest: PD):
        await self.delegation.send_illust(model, post_dest=post_dest)
        self.recorder.record_resp(model.id, post_dest.identifier)

    async def send_illusts(self, model: IllustMessagesModel,
                           *, post_dest: PD):
        await self.delegation.send_illusts(model, post_dest=post_dest)
        if len(model.messages) == 1:
            self.recorder.record_resp(model.messages[0].id, post_dest.identifier)

    def __getattr__(self, name: str):
        return getattr(self.delegation, name)



from typing import TypeVar, Optional, Any, AsyncGenerator, Collection

from beanie import Document, BulkWriter
from pymongo import IndexModel

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.model import Subscription, PostIdentifier
from ..source import MongoDataSource
from ..seq import SeqRepo
from ..utils.process_subscriber import process_subscriber

UID = TypeVar("UID")
GID = TypeVar("GID")

ID = PostIdentifier[UID, GID]


class SubscriptionDocument(Subscription[Any, Any], Document):
    class Settings:
        name = "subscription"
        indexes = [
            IndexModel([("subscriber.adapter", 1)]),
            IndexModel([("subscriber", 1), ("code", 1)], unique=True)
        ]


context.require(MongoDataSource).document_models.append(SubscriptionDocument)


@context.inject
@context.register_singleton()
class MongoSubscriptionRepo:
    mongo: MongoDataSource = Inject(MongoDataSource)
    seq_repo: SeqRepo = Inject(SeqRepo)

    async def get_by_subscriber(self, subscriber: ID) -> AsyncGenerator[Subscription, None]:
        subscriber = process_subscriber(subscriber)
        async for doc in SubscriptionDocument.find(SubscriptionDocument.subscriber == subscriber):
            yield doc

    async def get_by_adapter(self, adapter: str) -> AsyncGenerator[Subscription, None]:
        async for doc in SubscriptionDocument.find(SubscriptionDocument.subscriber.adapter == adapter):
            yield doc

    async def get_by_code(self, subscriber: ID, code: int) -> Optional[Subscription]:
        subscriber = process_subscriber(subscriber)
        return await SubscriptionDocument.find_one(SubscriptionDocument.subscriber == subscriber,
                                                   SubscriptionDocument.code == code)

    async def insert(self, subscription: Subscription):
        subscription.subscriber = process_subscriber(subscription.subscriber)
        subscription.code = await self.seq_repo.inc_and_get(f'subscription {subscription.subscriber}')
        await SubscriptionDocument.insert_one(SubscriptionDocument(**subscription.dict()))

    async def delete_one(self, subscriber: ID, code: int) -> Optional[Subscription]:
        # beanie不支持原子性的find_one_and_delete操作
        subscriber = process_subscriber(subscriber)
        query = {
            "code": code,
            "subscriber": process_subscriber(subscriber).dict()
        }
        result = await self.mongo.db.subscription.find_one_and_delete(query)
        if result:
            return Subscription.parse_obj(result)
        else:
            return None

    async def delete_many_by_subscriber(self, subscriber: ID) -> Collection[Subscription]:
        subscriber = process_subscriber(subscriber)

        old_doc = await SubscriptionDocument.find(
            SubscriptionDocument.subscriber == subscriber
        ).to_list()

        async with BulkWriter() as bw:
            for x in old_doc:
                await x.delete(bulk_writer=bw)

        return old_doc


__all__ = ("MongoSubscriptionRepo",)



__numerals = {'零': 0, '一': 1, '二': 2, '两': 2, '三': 3, '四': 4, '五': 5, '六': 6, '七': 7, '八': 8, '九': 9, '十': 10,
              '百': 100, '千': 1000, '万': 10000, '亿': 100000000}


def decode_chinese_integer(text: str) -> int:
    """
    将中文整数转换为int
    :param text: 中文整数
    :return: 对应int
    """
    ans = 0
    radix = 1
    for i in reversed(range(len(text))):
        if text[i] not in __numerals:
            raise ValueError()
        digit = __numerals[text[i]]
        if digit >= 10:
            if digit > radix:  # 成为新的基数
                radix = digit
                if i == 0:  # 若给定字符串省略了最前面的“一”，如十三、十五……
                    ans = ans + radix
            else:
                radix = radix * digit
        else:
            ans = ans + radix * digit

    return ans


def decode_integer(text: str) -> int:
    try:
        return int(text)
    except ValueError:
        pass

    return decode_chinese_integer(text)


__all__ = ("decode_integer", "decode_chinese_integer")



from typing import TypeVar, Sequence

from lazy import lazy
from nonebot import on_regex, Bot
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import Matcher
from nonebot.internal.params import Depends
from nonebot.typing import T_State

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.service.pixiv_account_binder import PixivAccountBinder
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from .base import RecordCommonHandler
from ..entry_handler import post_destination
from ..pkg_context import context
from ..utils import get_common_query_rule, get_count, get_post_dest

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.inject
@context.root.register_eager_singleton()
class RandomBookmarkHandler(RecordCommonHandler):
    binder = Inject(PixivAccountBinder)

    @classmethod
    def type(cls) -> str:
        return "random_bookmark"

    def enabled(self) -> bool:
        return self.conf.pixiv_random_bookmark_query_enabled

    @lazy
    def matcher(self):
        return on_regex("^来(.*)?张私家车$", rule=get_common_query_rule(), priority=5)

    async def on_match(self, bot: Bot, event: Event, state: T_State, matcher: Matcher,
                       post_dest: PostDestination[UID, GID] = Depends(post_destination)):
        await self.handle(count=get_count(state), post_dest=get_post_dest(bot, event))

    def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        pixiv_user_id = 0
        sender_user_id = post_dest.user_id

        if len(args) > 0:
            try:
                pixiv_user_id = int(args[0])
            except ValueError:
                raise BadRequestError(f"{args[0]}不是合法的ID")

        # 因为群组的订阅会把PostIdentifier的user_id抹去，所以这里必须传递sender_user_id
        return {"pixiv_user_id": pixiv_user_id, "sender_user_id": sender_user_id}

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, sender_user_id: UID,
                            pixiv_user_id: int = 0,
                            count: int = 1,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        if not pixiv_user_id and sender_user_id:
            pixiv_user_id = await self.binder.get_binding(post_dest.adapter, sender_user_id)

        if not pixiv_user_id:
            pixiv_user_id = self.conf.pixiv_random_bookmark_user_id

        if not pixiv_user_id:
            raise BadRequestError("无效的Pixiv账号，或未绑定Pixiv账号")

        illusts = await self.service.random_bookmark(pixiv_user_id, count=count)

        await self.post_illusts(illusts,
                                header="这是您点的私家车",
                                post_dest=post_dest)



from abc import ABC

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.handler.handler import Handler
from nonebot_plugin_pixivbot.service.pixiv_service import PixivService
from ..pkg_context import context


@context.inject
class WatchTaskHandler(Handler, ABC):
    service = Inject(PixivService)

    def parse_args(self, args, post_dest):
        raise RuntimeError("Please call handle_with_parsed_args() instead! ")



from typing import Optional, TypeVar, Protocol

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.enums import DataSourceType
from nonebot_plugin_pixivbot.model import PixivBinding

UID = TypeVar("UID")


class PixivBindingRepo(Protocol):
    async def get(self, adapter: str, user_id: UID) -> Optional[PixivBinding]:
        ...

    async def update(self, binding: PixivBinding):
        ...

    async def remove(self, adapter: str, user_id: UID) -> bool:
        ...


conf = context.require(Config)
if conf.pixiv_data_source == DataSourceType.mongo:
    from .mongo import MongoPixivBindingRepo

    context.bind(PixivBindingRepo, MongoPixivBindingRepo)
else:
    from .sql import SqlPixivBindingRepo

    context.bind(PixivBindingRepo, SqlPixivBindingRepo)

__all__ = ("PixivBindingRepo",)



from typing import Any

from beanie import Document
from pymongo import IndexModel

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.data.source import MongoDataSource


class MetaInfo(Document):
    key: str
    value: Any

    class Settings:
        name = "meta_info"
        indexes = [
            IndexModel([("key", 1)], unique=True),
        ]


context.require(MongoDataSource).document_models.append(MetaInfo)



from asyncio import create_task, gather
from typing import List, Type

from beanie import init_beanie, Document
from bson import CodecOptions
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase
from nonebot import logger
from pymongo import IndexModel
from pymongo.errors import OperationFailure

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.errors import DataSourceNotReadyError
from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.utils.lifecycler import on_shutdown, on_startup
from .migration import MongoMigrationManager


@context.inject
@context.register_eager_singleton()
class MongoDataSource:
    conf = Inject(Config)
    mongo_migration_mgr = Inject(MongoMigrationManager)
    app_db_version = 5

    def __init__(self):
        self._client = None
        self._db = None

        self.document_models: List[Type[Document]] = []

        on_startup(self.initialize, replay=True)
        on_shutdown(self.finalize)

    @property
    def client(self):
        return self._client

    @property
    def db(self):
        if self._db is not None:
            return self._db
        else:
            raise DataSourceNotReadyError()

    @staticmethod
    async def _raw_get_db_version(db: AsyncIOMotorDatabase) -> int:
        version = await db["meta_info"].find_one({"key": "db_version"})
        if version is None:
            await db["meta_info"].insert_one({"key": "db_version", "value": 1})
            return 1
        else:
            return version["value"]

    @staticmethod
    async def _raw_set_db_version(db: AsyncIOMotorDatabase, db_version: int):
        await db["meta_info"].update_one({"key": "db_version"},
                                         {"$set": {
                                             "value": db_version
                                         }},
                                         upsert=True)

    @staticmethod
    async def _ensure_ttl_index(db: AsyncIOMotorDatabase, coll_name: str, index: IndexModel):
        try:
            await db[coll_name].create_indexes([index])
        except OperationFailure:
            await db.command({
                "collMod": coll_name,
                "index": {
                    "keyPattern": index.document["key"],
                    "expireAfterSeconds": index.document["expireAfterSeconds"]
                }
            })
            logger.success(
                f"Index in {coll_name}: expireAfterSeconds changed to {index.document['expireAfterSeconds']}")

    async def initialize(self):
        client = AsyncIOMotorClient(self.conf.pixiv_mongo_conn_url)
        options = CodecOptions(tz_aware=True)
        db = client[self.conf.pixiv_mongo_database_name].with_options(options)

        # migrate
        db_version = await self._raw_get_db_version(db)
        await self.mongo_migration_mgr.perform_migration(db, db_version, self.app_db_version)
        await self._raw_set_db_version(db, self.app_db_version)

        # ensure ttl indexes (before init beanie)
        index_tasks = []
        for model in self.document_models:
            name = model.__name__

            settings = getattr(model, "Settings", None)
            if settings:
                settings_name = getattr(settings, "name")
                if settings_name:
                    name = settings_name

                settings_indexes = getattr(settings, "indexes")
                if settings_indexes:
                    for index in settings_indexes:
                        if isinstance(index, IndexModel) and index.document.get("expireAfterSeconds", None):
                            index_tasks.append(create_task(self._ensure_ttl_index(db, name, index)))
        await gather(*index_tasks)

        await init_beanie(database=db, document_models=self.document_models, allow_index_dropping=True)

        self._client = client
        self._db = db
        logger.success("MongoDataSource Initialization Succeed.")

    async def finalize(self):
        if self._client:
            self._client.close()
        self._client = None
        self._db = None


__all__ = ("MongoDataSource",)



from asyncio import create_task, gather
from typing import Optional, Sequence

from pydantic import BaseModel

from nonebot_plugin_pixivbot.model import Illust
from nonebot_plugin_pixivbot.model.message.illust_message import IllustMessageModel


class IllustMessagesModel(BaseModel):
    header: Optional[str] = None
    messages: Sequence[IllustMessageModel]

    @staticmethod
    async def from_illusts(illusts: Sequence[Illust], *,
                           header: Optional[str] = None,
                           number: Optional[int] = None) -> Optional["IllustMessagesModel"]:
        tasks = [
            create_task(
                IllustMessageModel.from_illust(x, number=number + i if number is not None else None)
            ) for i, x in enumerate(illusts)
        ]
        await gather(*tasks)

        messages = []
        for t in tasks:
            result = await t
            if result:
                messages.append(result)

        model = IllustMessagesModel(header=header, messages=messages)
        return model

    def flat(self):
        for x in self.messages:
            yield x.copy(update={"header": self.header})

    def flat_first(self) -> IllustMessageModel:
        for x in self.flat():
            return x



from datetime import datetime, timezone, timedelta
from typing import AsyncGenerator, Union, Optional, List

from nonebot import logger
from sqlalchemy import select, delete
from sqlalchemy.dialects.sqlite import insert

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.local_tag import LocalTagRepo
from nonebot_plugin_pixivbot.data.pixiv_repo import LazyIllust
from nonebot_plugin_pixivbot.data.pixiv_repo.errors import CacheExpiredError, NoSuchItemError
from nonebot_plugin_pixivbot.data.pixiv_repo.local_repo.sql_models import IllustDetailCache, UserDetailCache, \
    DownloadCache, IllustSetCache, IllustSetCacheIllust, UserSetCache, UserSetCacheUser
from nonebot_plugin_pixivbot.data.pixiv_repo.models import PixivRepoMetadata
from nonebot_plugin_pixivbot.data.source.sql import SqlDataSource
from nonebot_plugin_pixivbot.enums import RankingMode
from nonebot_plugin_pixivbot.model import Illust, User


def _handle_expires_in(metadata: PixivRepoMetadata, expires_in: int):
    if datetime.now(timezone.utc) - metadata.update_time >= timedelta(seconds=expires_in):
        raise CacheExpiredError(metadata)


def _extract_metadata(cache, is_set_cache):
    update_time = cache.update_time.replace(tzinfo=timezone.utc)

    if is_set_cache:
        metadata = PixivRepoMetadata(update_time=update_time, pages=cache.pages, next_qs=cache.next_qs)
    else:
        metadata = PixivRepoMetadata(update_time=update_time)
    return metadata


@context.inject
@context.register_singleton()
class SqlPixivRepo:
    conf: Config = Inject(Config)
    data_source: SqlDataSource = Inject(SqlDataSource)
    local_tag_repo: LocalTagRepo = Inject(LocalTagRepo)

    async def _get_illusts(self, cache_type: str,
                           key: dict,
                           expired_in: int,
                           offset: int = 0):
        session = self.data_source.session()
        stmt = (select(IllustSetCache)
                .where(IllustSetCache.cache_type == cache_type,
                       IllustSetCache.key == key))
        cache: Optional[IllustSetCache] = (await session.execute(stmt)).scalar_one_or_none()
        if cache is None:
            raise NoSuchItemError()

        metadata = _extract_metadata(cache, True)
        _handle_expires_in(metadata, expired_in)

        yield metadata.copy(update={"pages": 0})

        stmt = (select(IllustSetCacheIllust, IllustDetailCache)
                .where(IllustSetCacheIllust.cache_id == cache.id)
                .outerjoin(IllustDetailCache, IllustSetCacheIllust.illust_id == IllustDetailCache.illust_id)
                .order_by(IllustSetCacheIllust.rank)
                .offset(offset))

        total = 0
        broken = 0

        try:
            async for cache_illust, illust in await session.stream(stmt):
                cache_illust: IllustSetCacheIllust
                illust: Optional[IllustDetailCache]

                total += 1

                if illust is not None:
                    yield LazyIllust(illust.illust_id, Illust(**illust.illust))
                else:
                    yield LazyIllust(cache_illust.illust_id)
                    broken += 1

            yield metadata
        finally:
            logger.debug(f"[local] {total} got, illust_detail of {broken} are missed")

    async def _invalidate_illusts(self, cache_type: str, key: dict):
        session = self.data_source.session()
        stmt = (delete(IllustSetCache)
                .where(IllustSetCache.cache_type == cache_type,
                       IllustSetCache.key == key))
        await session.execute(stmt)
        await session.commit()

    async def _append_and_check_illusts(self, cache_type: str,
                                        key: dict,
                                        content: List[Union[Illust, LazyIllust]],
                                        metadata: PixivRepoMetadata,
                                        ranked: bool = False):
        session = self.data_source.session()

        async with session.begin_nested() as tx1:
            async with session.begin_nested() as tx2:
                stmt = (select(IllustSetCache)
                        .where(IllustSetCache.cache_type == cache_type,
                               IllustSetCache.key == key)
                        .limit(1))
                cache = (await session.execute(stmt)).scalar_one_or_none()
                if cache is None:
                    cache = IllustSetCache(cache_type=cache_type, key=key)
                    session.add(cache)

                cache.update_time = metadata.update_time
                cache.next_qs = metadata.next_qs
                cache.pages = metadata.pages

                # commit to get cache id
                await tx2.commit()

            if ranked:
                stmt = insert(IllustSetCacheIllust).values([
                    {"cache_id": cache.id, "illust_id": x.id, "rank": cache.size + i}
                    for i, x in enumerate(content)
                ]).on_conflict_do_nothing(index_elements=[
                    IllustSetCacheIllust.cache_id, IllustSetCacheIllust.illust_id
                ])
            else:
                stmt = insert(IllustSetCacheIllust).values([
                    {"cache_id": cache.id, "illust_id": x.id}
                    for i, x in enumerate(content)
                ]).on_conflict_do_nothing(index_elements=[
                    IllustSetCacheIllust.cache_id, IllustSetCacheIllust.illust_id
                ])

            row_count = (await session.execute(stmt)).rowcount
            if ranked:
                cache.size += row_count

            await tx1.commit()

        # insert illust detail
        values = []
        for x in content:
            if isinstance(x, LazyIllust):
                if not x.loaded:
                    continue
                x = x.content
            values.append(dict(illust_id=x.id, illust=x.dict(), update_time=metadata.update_time))
        if len(values) != 0:
            stmt = insert(IllustDetailCache).values(values)
            await session.execute(stmt)
            await session.commit()

        return row_count != len(content)

    async def _get_users(self, cache_type: str,
                         key: dict,
                         expired_in: int,
                         offset: int = 0):
        session = self.data_source.session()
        stmt = (select(UserSetCache)
                .where(UserSetCache.cache_type == cache_type,
                       UserSetCache.key == key))
        cache = (await session.execute(stmt)).scalar_one_or_none()
        if cache is None:
            raise NoSuchItemError()

        metadata = _extract_metadata(cache, True)
        _handle_expires_in(metadata, expired_in)

        yield metadata.copy(update={"pages": 0})

        stmt = (select(UserSetCache, UserDetailCache)
                .where(UserSetCacheUser.cache_id == cache.id)
                .outerjoin(UserDetailCache, UserSetCacheUser.user_id == UserDetailCache.user_id)
                .offset(offset))

        async for cache_user, user in await session.stream(stmt):
            if user is not None:
                yield User(**user.user)
            else:
                yield User(id=cache_user.user_id, name="", account="")

        yield metadata

    async def _invalidate_users(self, cache_type: str, key: dict):
        session = self.data_source.session()
        stmt = (delete(UserSetCache)
                .where(UserSetCache.cache_type == cache_type,
                       UserSetCache.key == key))
        await session.execute(stmt)
        await session.commit()

    async def _append_and_check_users(self, cache_type: str,
                                      key: dict,
                                      content: List[User],
                                      metadata: PixivRepoMetadata):
        session = self.data_source.session()

        async with session.begin_nested() as tx1:
            async with session.begin_nested() as tx2:
                stmt = (select(UserSetCache)
                        .where(UserSetCache.cache_type == cache_type,
                               UserSetCache.key == key)
                        .limit(1))
                cache = (await session.execute(stmt)).scalar_one_or_none()
                if cache is None:
                    cache = UserSetCache(cache_type=cache_type, key=key)
                    session.add(cache)

                cache.update_time = metadata.update_time
                cache.next_qs = metadata.next_qs
                cache.pages = metadata.pages

                # commit to get cache id
                await tx2.commit()

            stmt = insert(UserSetCacheUser).values([
                {"cache_id": cache.id, "user_id": x.id}
                for i, x in enumerate(content)
            ]).on_conflict_do_nothing(index_elements=[
                UserSetCacheUser.cache_id, UserSetCacheUser.user_id
            ])

            row_count = (await session.execute(stmt)).rowcount
            await tx1.commit()

        # insert user detail
        if len(content) != 0:
            values = [dict(user_id=x.id, user=x.dict(), update_time=metadata.update_time) for x in content]
            stmt = insert(UserDetailCache).values(values)
            await session.execute(stmt)
            await session.commit()

        return row_count != len(content)

    # ================ illust_detail ================
    async def illust_detail(self, illust_id: int) \
            -> AsyncGenerator[Union[Illust, PixivRepoMetadata], None]:
        logger.debug(f"[local] illust_detail {illust_id}")

        session = self.data_source.session()
        stmt = select(IllustDetailCache).where(IllustDetailCache.illust_id == illust_id).limit(1)
        cache = (await session.execute(stmt)).scalar_one_or_none()

        if cache is not None:
            metadata = _extract_metadata(cache, False)
            _handle_expires_in(metadata, self.conf.pixiv_illust_detail_cache_expires_in)

            yield metadata
            yield Illust(**cache.illust)
        else:
            raise NoSuchItemError()

    async def update_illust_detail(self, illust: Illust, metadata: PixivRepoMetadata):
        logger.debug(f"[local] update illust_detail {illust.id} {metadata}")

        session = self.data_source.session()
        stmt = (insert(IllustDetailCache)
                .values(illust_id=illust.id, illust=illust.dict(), update_time=metadata.update_time))

        await session.execute(stmt)
        await session.commit()

        if self.conf.pixiv_tag_translation_enabled:
            await self.local_tag_repo.update_from_illusts([illust])

    # ================ user_detail ================
    async def user_detail(self, user_id: int) \
            -> AsyncGenerator[Union[User, PixivRepoMetadata], None]:
        logger.debug(f"[local] user_detail {user_id}")

        session = self.data_source.session()
        stmt = select(UserDetailCache).where(UserDetailCache.user_id == user_id).limit(1)
        cache = (await session.execute(stmt)).scalar_one_or_none()

        if cache is not None:
            metadata = _extract_metadata(cache, False)
            _handle_expires_in(metadata, self.conf.pixiv_user_detail_cache_expires_in)

            yield metadata
            yield User(**cache.user)
        else:
            raise NoSuchItemError()

    async def update_user_detail(self, user: User, metadata: PixivRepoMetadata):
        logger.debug(f"[local] update user_detail {user.id} {metadata}")

        session = self.data_source.session()
        stmt = (insert(UserDetailCache)
                .values(user_id=user.id, user=user.dict(), update_time=metadata.update_time))

        await session.execute(stmt)
        await session.commit()

    # ================ image ================
    async def image(self, illust: Illust) -> AsyncGenerator[Union[bytes, PixivRepoMetadata], None]:
        logger.debug(f"[local] image {illust.id}")

        session = self.data_source.session()
        stmt = select(DownloadCache).where(DownloadCache.illust_id == illust.id).limit(1)
        cache = (await session.execute(stmt)).scalar_one_or_none()

        if cache is not None:
            metadata = _extract_metadata(cache, False)
            _handle_expires_in(metadata, self.conf.pixiv_download_cache_expires_in)

            yield metadata
            yield cache.content
        else:
            raise NoSuchItemError()

    async def update_image(self, illust_id: int, content: bytes,
                           metadata: PixivRepoMetadata):
        logger.debug(f"[local] update image {illust_id} {metadata}")

        session = self.data_source.session()
        stmt = (insert(DownloadCache)
                .values(illust_id=illust_id, content=content, update_time=metadata.update_time))

        await session.execute(stmt)
        await session.commit()

    # ================ illust_ranking ================
    def illust_ranking(self, mode: Union[str, RankingMode], *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        if isinstance(mode, str):
            mode = RankingMode[mode]

        logger.debug(f"[local] illust_ranking {mode}")

        return self._get_illusts("illust_ranking", {"mode": mode},
                                 expired_in=self.conf.pixiv_illust_ranking_cache_expires_in, offset=offset)

    async def invalidate_illust_ranking(self, mode: RankingMode):
        logger.debug(f"[local] invalidate illust_ranking")
        await self._invalidate_illusts("illust_ranking", {"mode": mode})

    async def append_illust_ranking(self, mode: RankingMode, content: List[Union[Illust, LazyIllust]],
                                    metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append illust_ranking {mode} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts("illust_ranking", {"mode": mode},
                                                    content=content, metadata=metadata,
                                                    ranked=True)

    # ================ search_illust ================
    def search_illust(self, word: str, *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[local] search_illust {word}")
        return self._get_illusts("search_illust", {"word": word},
                                 expired_in=self.conf.pixiv_search_illust_cache_expires_in, offset=offset)

    async def invalidate_search_illust(self, word: str):
        logger.debug(f"[local] invalidate search_illust {word}")
        await self._invalidate_illusts("search_illust", {"word": word})

    async def append_search_illust(self, word: str, content: List[Union[Illust, LazyIllust]],
                                   metadata: PixivRepoMetadata) -> bool:
        # 返回值表示content中是否有已经存在于集合的文档，下同
        logger.debug(f"[local] append search_illust {word} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts("search_illust", {"word": word},
                                                    content=content, metadata=metadata)

    # ================ user_illusts ================
    def user_illusts(self, user_id: int, *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[local] user_illusts {user_id}")
        return self._get_illusts("user_illusts", {"user_id": user_id},
                                 expired_in=self.conf.pixiv_user_illusts_cache_expires_in, offset=offset)

    async def invalidate_user_illusts(self, user_id: int):
        logger.debug(f"[local] invalidate user_illusts {user_id}")
        await self._invalidate_illusts("user_illusts", {"user_id": user_id})

    async def append_user_illusts(self, user_id: int,
                                  content: List[Union[Illust, LazyIllust]],
                                  metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append user_illusts {user_id} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts("user_illusts", {"user_id": user_id},
                                                    content=content, metadata=metadata)

    # ================ user_bookmarks ================
    def user_bookmarks(self, user_id: int = 0, *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[local] user_bookmarks {user_id}")
        return self._get_illusts("user_bookmarks", {"user_id": user_id},
                                 expired_in=self.conf.pixiv_user_bookmarks_cache_expires_in, offset=offset)

    async def invalidate_user_bookmarks(self, user_id: int):
        logger.debug(f"[local] invalidate user_bookmarks {user_id}")
        await self._invalidate_illusts("user_bookmarks", {"user_id": user_id})

    async def append_user_bookmarks(self, user_id: int,
                                    content: List[Union[Illust, LazyIllust]],
                                    metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append user_bookmarks {user_id} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts("user_bookmarks", {"user_id": user_id},
                                                    content=content, metadata=metadata)

    # ================ recommended_illusts ================
    def recommended_illusts(self, *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[local] recommended_illusts")
        return self._get_illusts("other", {"type": "recommended_illusts"},
                                 expired_in=self.conf.pixiv_other_cache_expires_in, offset=offset)

    async def invalidate_recommended_illusts(self):
        logger.debug(f"[local] invalidate recommended_illusts")
        await self._invalidate_illusts("other", {"type": "recommended_illusts"})

    async def append_recommended_illusts(self, content: List[Union[Illust, LazyIllust]],
                                         metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append recommended_illusts "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts("other", {"type": "recommended_illusts"},
                                                    content=content, metadata=metadata)

    # ================ related_illusts ================
    def related_illusts(self, illust_id: int, *, offset: int = 0) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[local] related_illusts {illust_id}")
        return self._get_illusts("related_illusts", {"original_illust_id": illust_id},
                                 expired_in=self.conf.pixiv_related_illusts_cache_expires_in, offset=offset)

    async def invalidate_related_illusts(self, illust_id: int):
        logger.debug(f"[local] invalidate related_illusts")
        await self._invalidate_illusts("related_illusts", {"original_illust_id": illust_id})

    async def append_related_illusts(self, illust_id: int, content: List[Union[Illust, LazyIllust]],
                                     metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append related_illusts {illust_id} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_illusts("related_illusts", {"original_illust_id": illust_id},
                                                    content=content, metadata=metadata)

    # ================ search_user ================
    def search_user(self, word: str, *, offset: int = 0) \
            -> AsyncGenerator[Union[User, PixivRepoMetadata], None]:
        logger.debug(f"[local] search_user {word}")
        return self._get_users("search_user", {"word": word},
                               expired_in=self.conf.pixiv_search_user_cache_expires_in, offset=offset)

    async def invalidate_search_user(self, word: str):
        logger.debug(f"[local] invalidate search_user {word}")
        await self._invalidate_users("search_user", {"word": word})

    async def append_search_user(self, word: str, content: List[User],
                                 metadata: PixivRepoMetadata) -> bool:
        logger.debug(f"[local] append search_user {word} "
                     f"({len(content)} items) "
                     f"{metadata}")
        return await self._append_and_check_users("search_user", {"word": word},
                                                  content=content, metadata=metadata)

    async def invalidate_all(self):
        logger.debug(f"[local] invalidate_all")

        session = self.data_source.session()
        await session.execute(delete(IllustSetCache))
        await session.execute(delete(UserSetCache))
        await session.execute(delete(IllustDetailCache))
        await session.execute(delete(UserDetailCache))
        await session.execute(delete(DownloadCache))



from typing import TypeVar, Sequence

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.pixiv_repo import PixivRepo
from nonebot_plugin_pixivbot.handler.interceptor.permission_interceptor import SuperuserInterceptor
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from .command import SubCommandHandler, CommandHandler
from ..pkg_context import context

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.inject
@context.require(CommandHandler).sub_command("invalidate_cache")
class InvalidateCacheHandler(SubCommandHandler):
    repo = Inject(PixivRepo)

    def __init__(self):
        super().__init__()
        self.add_interceptor(context.require(SuperuserInterceptor))

    @classmethod
    def type(cls) -> str:
        return "invalidate_cache"

    def enabled(self) -> bool:
        return True

    def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        return {}

    async def actual_handle(self, *, post_dest: PostDestination[UID, GID],
                            silently: bool = False, **kwargs):
        await self.repo.invalidate_cache()
        await self.post_plain_text(message="ok", post_dest=post_dest)



from datetime import datetime, timezone
from typing import Optional

from pydantic import BaseModel


class PixivRepoMetadata(BaseModel):
    update_time: datetime = datetime.now(timezone.utc)
    pages: Optional[int]
    next_qs: Optional[dict]

    def __str__(self):
        return '(' + ', '.join(map(lambda kv: kv[0] + '=' + str(kv[1]), self.dict(exclude_none=True).items())) + ')'


__all__ = ("PixivRepoMetadata",)



from collections import OrderedDict
from collections.abc import MutableMapping
from dataclasses import dataclass
from datetime import datetime, timezone
from functools import total_ordering
from heapq import heappop, heapify, heappush
from typing import TypeVar, Generic, Iterator, List, Optional, Callable

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


@dataclass(frozen=True)
@total_ordering
class ExpiresNode(Generic[_KT, _VT]):
    key: _KT
    value: _VT
    expires_time: datetime

    def __le__(self, other: "ExpiresNode"):
        return self.expires_time < other.expires_time


class ExpiresLruDict(MutableMapping[_KT, _VT], Generic[_KT, _VT]):
    def __init__(self, maxsize: int, on_cleanup: Optional[Callable[[_KT, _VT], None]] = None):
        self.maxsize = maxsize
        self.on_cleanup = on_cleanup

        self._data = OrderedDict[_KT, ExpiresNode[_KT, _VT]]()
        self._expires_heap: List[ExpiresNode[_KT, _VT]] = []

    def _collate(self, ensure_size_for_put: bool):
        now = datetime.now(timezone.utc)
        while len(self._expires_heap) > 0 and self._expires_heap[0].expires_time <= now:
            node = self._expires_heap[0]
            if self._data.get(node.key, None) is node:
                del self._data[node.key]
                self.on_cleanup and self.on_cleanup(node.key, node.value)

            heappop(self._expires_heap)

        if ensure_size_for_put and len(self._data) == self.maxsize:
            key, node = self._data.popitem(last=False)
            self.on_cleanup and self.on_cleanup(key, node.value)
            # we will not remove items from expires_heap
            # until its size reach the threshold (checked on add)

    def _collate_expires_heap(self):
        threshold = 2 * self.maxsize - 1

        if len(self._expires_heap) > threshold:
            # rebuild _expires_heap
            self._expires_heap = []
            for node in self._data.values():
                self._expires_heap.append(node)
            heapify(self._expires_heap)

    def add(self, key: _KT, value: _VT, expires_time: datetime):
        self._collate(True)

        if self.__contains__(key):
            raise KeyError(key)

        node = ExpiresNode(key, value, expires_time)
        self._data.__setitem__(key, node)
        heappush(self._expires_heap, node)

        self._collate_expires_heap()

    def __setitem__(self, key: _KT, value: _VT) -> None:
        self._collate(False)
        if key not in self._data:
            raise KeyError(key)

        node = self._data.__getitem__(key)
        self._data[key] = ExpiresNode(key, value, node.expires_time)

    def __delitem__(self, key: _KT) -> None:
        self._collate(False)

        # we will not remove items from expires_heap
        # until its size reach the threshold (checked on add)
        return self._data.__delitem__(key)

    def __getitem__(self, key: _KT) -> _VT:
        self._collate(False)
        return self._data.__getitem__(key).value

    def __len__(self) -> int:
        self._collate(False)
        return self._data.__len__()

    def __iter__(self) -> Iterator[_KT]:
        self._collate(False)
        for node in self._data.__iter__():
            yield node



import datetime
import typing

from pydantic import *

from .tag import Tag
from .user import User


class Illust(BaseModel):
    class ImageUrls(BaseModel):
        square_medium: str
        medium: str
        large: str

    class MetaSinglePage(BaseModel):
        original_image_url: typing.Optional[str] = None

    class MetaPage(BaseModel):
        class ImageUrls(BaseModel):
            square_medium: str
            medium: str
            large: str
            original: str

        image_urls: ImageUrls

    id: int
    title: str
    type: str
    image_urls: ImageUrls
    caption: str
    user: User
    tags: typing.List[Tag]
    create_date: datetime.datetime
    page_count: int
    meta_single_page: MetaSinglePage
    meta_pages: typing.List[MetaPage]
    total_view: int
    total_bookmarks: int

    def has_tag(self, tag: typing.Union[str, Tag]) -> bool:
        if isinstance(tag, Tag):
            for x in self.tags:
                if x == tag:
                    return True
            return False
        else:
            for x in self.tags:
                if x.name == tag or x.translated_name == tag:
                    return True
            return False

    def has_tags(self, tags: typing.List[typing.Union[str, Tag]]) -> bool:
        for tag in tags:
            if self.has_tag(tag):
                return True
        return False


__all__ = ("Illust",)



from typing import List

from pydantic import BaseModel

from nonebot_plugin_pixivbot.model import User, Illust


class UserPreview(BaseModel):
    user: User
    illusts: List[Illust]

    class Config:
        extra = "ignore"



from typing import TypeVar, Sequence

from lazy import lazy
from nonebot import on_regex, Bot
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import Matcher
from nonebot.internal.params import Depends
from nonebot.typing import T_State

from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from .base import RecordCommonHandler
from ..entry_handler import post_destination
from ..pkg_context import context
from ..utils import get_common_query_rule, get_count

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.root.register_eager_singleton()
class RandomIllustHandler(RecordCommonHandler):
    @classmethod
    def type(cls) -> str:
        return "random_illust"

    def enabled(self) -> bool:
        return self.conf.pixiv_random_illust_query_enabled

    @lazy
    def matcher(self):
        return on_regex("^来(.*)?张(.+)图$", rule=get_common_query_rule(), priority=5)

    async def on_match(self, bot: Bot, event: Event, state: T_State, matcher: Matcher,
                       post_dest: PostDestination[UID, GID] = Depends(post_destination)):
        word = state["_matched_groups"][1]
        await self.handle(word, count=get_count(state), post_dest=post_dest)

    def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        return {"word": args[0]}

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, word: str,
                            count: int = 1,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        illusts = await self.service.random_illust(word, count=count)

        await self.post_illusts(illusts,
                                header=f"这是您点的{word}图",
                                post_dest=post_dest)



from motor.motor_asyncio import AsyncIOMotorDatabase

from nonebot_plugin_pixivbot import context
from .mongo_migration import MongoMigration
from .mongo_migration_manager import MongoMigrationManager


@context.require(MongoMigrationManager).register
class MongoV3ToV4(MongoMigration):
    from_db_version = 3
    to_db_version = 4

    async def migrate(self, db: AsyncIOMotorDatabase):
        await self.migrate_subscription(db)

    async def migrate_subscription(self, db: AsyncIOMotorDatabase):
        try:
            await db["subscription"].drop_index("adapter_1_group_id_1_type_1")
        except:
            pass

        try:
            await db["subscription"].drop_index("adapter_1_user_id_1_type_1")
        except:
            pass

        await db["subscription"].update_many(
            {},
            [
                {
                    "$set": {
                        "subscriber": {
                            "$let": {
                                "vars": {
                                    "adapter": "$$ROOT.adapter",
                                    "user_id": "$$ROOT.user_id",
                                    "group_id": "$$ROOT.group_id",
                                },
                                "in": {
                                    "adapter": {
                                        "$cond": {
                                            "if": {"$ne": [{"$type": "$$adapter"}, "missing"]},
                                            "then": "$$adapter",
                                            "else": None
                                        },
                                    },
                                    "user_id": {
                                        "$cond": {
                                            "if": {"$ne": [{"$type": "$$user_id"}, "missing"]},
                                            "then": "$$user_id",
                                            "else": None
                                        },
                                    },
                                    "group_id": {
                                        "$cond": {
                                            "if": {"$ne": [{"$type": "$$group_id"}, "missing"]},
                                            "then": "$$group_id",
                                            "else": None
                                        },
                                    }
                                }
                            }
                        },
                        "tz": "Asia/Shanghai"
                    }
                },
                {
                    "$unset": [
                        "adapter",
                        "user_id",
                        "group_id"
                    ]
                }
            ]
        )



from .illust import IllustHandler
from .more import MoreHandler
from .random_bookmark import RandomBookmarkHandler
from .random_illust import RandomIllustHandler
from .random_recommended_illust import RandomRecommendedIllustHandler
from .random_related_illust import RandomRelatedIllustHandler
from .random_user_illust import RandomUserIllustHandler
from .ranking import RankingHandler

__all__ = ("RandomBookmarkHandler", "RandomRecommendedIllustHandler",
           "RankingHandler", "RandomIllustHandler",
           "RandomUserIllustHandler", "IllustHandler",
           "MoreHandler", "RandomRelatedIllustHandler")



from motor.motor_asyncio import AsyncIOMotorDatabase

from nonebot_plugin_pixivbot import context
from .mongo_migration import MongoMigration
from .mongo_migration_manager import MongoMigrationManager


@context.require(MongoMigrationManager).register
class MongoV2ToV3(MongoMigration):
    from_db_version = 2
    to_db_version = 3

    async def migrate(self, db: AsyncIOMotorDatabase):
        await db.download_cache.delete_many({})
        await db.illust_detail_cache.delete_many({})
        await db.user_detail_cache.delete_many({})
        await db.illust_ranking_cache.delete_many({})
        await db.search_illust_cache.delete_many({})
        await db.search_user_cache.delete_many({})
        await db.user_illusts_cache.delete_many({})
        await db.user_bookmarks_cache.delete_many({})
        await db.other_cache.delete_many({})



from typing import TypeVar, Generic, Sequence

from lazy import lazy
from nonebot import on_regex, Bot
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import Matcher
from nonebot.internal.params import Depends
from nonebot.typing import T_State

from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from .base import CommonHandler
from ..entry_handler import post_destination
from ..pkg_context import context
from ..utils import get_common_query_rule

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.root.register_eager_singleton()
class IllustHandler(CommonHandler, Generic[UID, GID]):
    @classmethod
    def type(cls) -> str:
        return "illust"

    def enabled(self) -> bool:
        return self.conf.pixiv_illust_query_enabled

    @lazy
    def matcher(self):
        return on_regex(r"^看看图\s*([1-9][0-9]*)$", rule=get_common_query_rule(), priority=5)

    async def on_match(self, bot: Bot, event: Event, state: T_State, matcher: Matcher,
                       post_dest: PostDestination[UID, GID] = Depends(post_destination)):
        illust_id = state["_matched_groups"][0]
        await self.handle(illust_id, post_dest=post_dest)

    def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        try:
            return {"illust_id": int(args[0])}
        except ValueError:
            raise BadRequestError(f"{args[0]}不是合法的插画ID")

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, illust_id: int,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False, **kwargs):
        illust = await self.service.illust_detail(illust_id)
        await self.post_illust(illust, post_dest=post_dest)



from abc import ABC, abstractmethod
from typing import TypeVar, Generic

from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.model.message import IllustMessageModel, IllustMessagesModel
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.protocol_dep.protocol_dep import ProtocolDep, ProtocolDepManager

UID = TypeVar("UID")
GID = TypeVar("GID")

PD = PostDestination[UID, GID]


class Postman(ProtocolDep, ABC, Generic[UID, GID]):
    @abstractmethod
    async def send_plain_text(self, message: str,
                              *, post_dest: PD):
        raise NotImplementedError()

    @abstractmethod
    async def send_illust(self, model: IllustMessageModel,
                          *, post_dest: PD):
        raise NotImplementedError()

    @abstractmethod
    async def send_illusts(self, model: IllustMessagesModel,
                           *, post_dest: PD):
        raise NotImplementedError()


@context.register_singleton()
class PostmanManager(ProtocolDepManager[Postman]):
    async def send_plain_text(self, message: str,
                              *, post_dest: PD):
        return await self[post_dest.adapter].send_plain_text(message, post_dest=post_dest)

    async def send_illust(self, model: IllustMessageModel,
                          *, post_dest: PD):
        return await self[post_dest.adapter].send_illust(model, post_dest=post_dest)

    async def send_illusts(self, model: IllustMessagesModel,
                           *, post_dest: PD):
        return await self[post_dest.adapter].send_illusts(model, post_dest=post_dest)


__all__ = ("Postman", "PostmanManager")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nonebot_plugin_pixivbot',
 'nonebot_plugin_pixivbot.data',
 'nonebot_plugin_pixivbot.data.local_tag',
 'nonebot_plugin_pixivbot.data.pixiv_binding',
 'nonebot_plugin_pixivbot.data.pixiv_repo',
 'nonebot_plugin_pixivbot.data.pixiv_repo.local_repo',
 'nonebot_plugin_pixivbot.data.seq',
 'nonebot_plugin_pixivbot.data.source',
 'nonebot_plugin_pixivbot.data.source.mongo',
 'nonebot_plugin_pixivbot.data.source.mongo.migration',
 'nonebot_plugin_pixivbot.data.subscription',
 'nonebot_plugin_pixivbot.data.utils',
 'nonebot_plugin_pixivbot.data.watch_task',
 'nonebot_plugin_pixivbot.handler',
 'nonebot_plugin_pixivbot.handler.command',
 'nonebot_plugin_pixivbot.handler.common',
 'nonebot_plugin_pixivbot.handler.interceptor',
 'nonebot_plugin_pixivbot.handler.sniffer',
 'nonebot_plugin_pixivbot.handler.watch',
 'nonebot_plugin_pixivbot.model',
 'nonebot_plugin_pixivbot.model.message',
 'nonebot_plugin_pixivbot.model.old',
 'nonebot_plugin_pixivbot.protocol_dep',
 'nonebot_plugin_pixivbot.service',
 'nonebot_plugin_pixivbot.utils']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.2.0,<10.0.0',
 'PixivPy-Async>=1.2.14,<2.0.0',
 'aiohttp-socks>=0.7.1,<0.8.0',
 'beanie>=1.12.0,<2.0.0',
 'frozendict>=2.3.4,<3.0.0',
 'lazy>=1.4,<2.0',
 'motor>=3.0.0,<4.0.0',
 'nonebot2>=2.0.0rc1,<3.0.0',
 'nonebot_plugin_apscheduler>=0.2.0,<0.3.0',
 'numpy>=1.23.1,<2.0.0',
 'sqlalchemy[asyncio]>=1.4.42,<2.0.0',
 'tzlocal>=4.2,<5.0']

extras_require = \
{'kook': ['nonebot-plugin-pixivbot-kook>=1.4.0,<2.0.0'],
 'onebot': ['nonebot-plugin-pixivbot-onebot-v11>=1.4.0,<2.0.0']}

setup_kwargs = {
    'name': 'nonebot-plugin-pixivbot',
    'version': '1.5.0b1',
    'description': 'Nonebot Plugin PixivBot',
    'long_description': 'nonebot_plugin_pixivbot\n=====\n\nPixivBot中协议无关的通用部分。目前适配协议：\n\n- [Onebot V11](https://github.com/ssttkkl/nonebot-plugin-pixivbot-onebot-v11)：`pip install nonebot-plugin-pixivbot[onebot]`\n- [KOOK / 开黑啦](https://github.com/ssttkkl/nonebot-plugin-pixivbot-kook)：`pip install nonebot-plugin-pixivbot[kook]`\n\n没有找到需要的协议？欢迎适配。[适配指南](https://github.com/ssttkkl/nonebot-plugin-pixivbot/wiki/%E9%80%82%E9%85%8D%E6%8C%87%E5%8D%97)\n\n开箱即用的Docker镜像：[ssttkkl/PixivBot](https://github.com/ssttkkl/PixivBot)\n\n## 触发语句\n\n### 普通语句\n\n所有数字参数均支持中文数字和罗马数字。\n\n- **看看<类型>榜<范围>**：查看pixiv榜单（<类型>可省略，<范围>应为a-b或a）\n  - 示例：看看榜、看看日榜、看看榜1-5、看看月榜一\n- **来<数量>张图**：从推荐插画随机抽选一张插画（<数量>可省略，下同）\n  - 示例：来张图、来五张图\n- **来<数量>张<关键字>图**：搜索关键字，从搜索结果随机抽选一张插画\n  - 示例：来张初音ミク图、来五张初音ミク图\n  - 注：默认开启关键字翻译功能。Bot会在平时的数据爬取时记录各个Tag的中文翻译。在搜索时，若关键字的日文翻译存在，则使用日文翻译代替关键字进行搜索。\n- **来<数量>张<用户>老师的图**：搜索用户，从插画列表中随机抽选一张插画\n  - 示例：来张Rella老师的图、来五张Rella老师的图\n- **看看图<插画ID>**：查看ID对应的插画\n  - 示例：看看图114514\n- **来<数量>张私家车**：从书签中随机抽选一张插画（发送者需绑定Pixiv账号，或者在配置中指定默认Pixiv账号）\n  - 示例：来张私家车、来五张私家车\n- **还要**：重复上一次请求\n- **不够色**：获取上一张插画的相关插画\n\n### 命令语句\n\n- **/pixivbot schedule \\<type\\> \\<schedule\\> [..args]**：为本群（本用户）订阅类型为<type>的定时推送功能，时间满足<schedule>时进行推送\n    - \\<type\\>：可选值有ranking, random_bookmark, random_recommended_illust, random_illust, random_user_illust\n    - \\<schedule\\>：有三种格式，*00:30\\*x*为每隔30分钟进行一次推送，*12:00*为每天12:00进行一次推送，*00:10+00:30\\*x*为从今天00:10开始每隔30分钟进行一次推送（开始时间若是一个过去的时间点，则从下一个开始推送的时间点进行推送）\n    - [..args]：\n      - \\<type\\>为ranking时，接受\\<mode\\> \\<range\\>\n      - \\<type\\>为random_bookmark时，接受\\<pixiv_user_id\\>（可空）\n      - \\<type\\>为random_illust时，接受\\<word\\>\n      - \\<type\\>为random_user_illust时，接受\\<user\\>\n- **/pixivbot schedule**：查看本群（本用户）的所有定时推送订阅\n- **/pixivbot unschedule <type>**：取消本群（本用户）的定时推送订阅\n- **/pixivbot watch \\<type\\> [..args]**：为本群（本用户）订阅类型为<type>的更新推送功能\n    - \\<type\\>：可选值有user_illusts, following_illusts\n    - [..args]：\n      - \\<type\\>为user_illusts时，接受\\<pixiv_user_id\\>/\\<user\\>\n      - \\<type\\>为following_illusts时，接受\\<pixiv_user_id\\>/\\<user\\>（可空）\n- **/pixivbot watch**：查看本群（本用户）的所有更新推送订阅\n- **/pixivbot unwatch \\<type\\> [..args]**：取消本群（本用户）的更新推送订阅\n- **/pixivbot bind \\<pixiv_user_id\\>**：绑定Pixiv账号（用于随机书签功能）\n- **/pixivbot unbind**：解绑Pixiv账号\n- **/pixivbot invalidate_cache**：清除缓存（只有超级用户能够发送此命令）\n- **/pixivbot**、**/pixivbot help**：查看帮助\n\n\n## 环境配置\n\n事前准备：登录pixiv账号并获取refresh_token。（参考：[@ZipFile Pixiv OAuth Flow](https://gist.github.com/ZipFile/c9ebedb224406f4f11845ab700124362)）\n\n1. 参考[安装 | NoneBot](https://v2.nonebot.dev/docs/start/installation)安装NoneBot；\n2. 参考[创建项目 | NoneBot](https://v2.nonebot.dev/docs/tutorial/create-project)创建一个NoneBot实例；\n3. 使用`pip install nonebot-plugin-pixivbot[xxx]`安装特定协议版本的插件；\n4. 修改pyproject.toml，启用插件（`plugins = [..., "nonebot_plugin_pixivbot"]`）；\n5. 安装MongoDB，并创建一个数据库及用户供应用使用；\n6. 在.env.prod中修改配置（参考下方）；\n\n## 配置\n\n最小配置：\n```\npixiv_refresh_token=  # 前面获取的REFRESH_TOKEN\npixiv_mongo_conn_url=  # MongoDB连接URL，格式：mongodb://<用户名>:<密码>@<主机>:<端口>/<数据库>\npixiv_mongo_database_name=  # 连接的MongoDB数据库\n```\n\n完整配置（除最小配置出现的配置项以外都是可选项，给出的是默认值）（NoneBot配置项这里不列出，参考[配置 | NoneBot](https://v2.nonebot.dev/docs/tutorial/configuration#%E8%AF%A6%E7%BB%86%E9%85%8D%E7%BD%AE%E9%A1%B9)）：\n\n```\nsuperuser=[]  # 能够发送超级命令的用户（JSON数组，格式为["onebot:123456", "kaiheila:1919810"]，下同）\nblocklist=[]  # Bot不响应的用户，可以避免Bot之间相互调用（JSON数组）\n\npixiv_refresh_token=  # 前面获取的REFRESH_TOKEN\npixiv_mongo_conn_url=  # MongoDB连接URL，格式：mongodb://<用户名>:<密码>@<主机>:<端口>/<数据库>\npixiv_mongo_database_name=  # 连接的MongoDB数据库\npixiv_proxy=None  # 代理URL\npixiv_query_timeout=60  # 查询超时（单位：秒）\npixiv_loading_prompt_delayed_time=5  # 加载提示消息的延迟时间（“努力加载中”的消息会在请求发出多少秒后发出）（单位：秒）\npixiv_simultaneous_query=8  # 向Pixiv查询的并发数\n\n# 缓存过期时间（单位：秒）\npixiv_download_cache_expires_in = 3600 * 24 * 7\npixiv_illust_detail_cache_expires_in = 3600 * 24 * 7\npixiv_user_detail_cache_expires_in = 3600 * 24 * 7\npixiv_illust_ranking_cache_expires_in = 3600 * 6\npixiv_search_illust_cache_expires_in = 3600 * 24\npixiv_search_user_cache_expires_in = 3600 * 24\npixiv_user_illusts_cache_expires_in = 3600 * 24\npixiv_user_bookmarks_cache_expires_in = 3600 * 24\npixiv_related_illusts_cache_expires_in = 3600 * 24\npixiv_other_cache_expires_in = 3600 * 6\n\npixiv_block_tags=[]  # 当插画含有指定tag时会被过滤\npixiv_block_action=no_image  # 过滤时的动作，可选值：no_image(不显示插画，回复插画信息), completely_block(只回复过滤提示), no_reply(无回复)\n\npixiv_download_quantity=original  # 插画下载品质，可选值：original, square_medium, medium, large\npixiv_download_custom_domain=None  # 使用反向代理下载插画的域名\n\npixiv_compression_enabled=False  # 启用插画压缩\npixiv_compression_max_size=None  # 插画压缩最大尺寸\npixiv_compression_quantity=None  # 插画压缩品质（0到100）\n\npixiv_query_to_me_only=False  # 只响应关于Bot的查询\npixiv_command_to_me_only=False  # 只响应关于Bot的命令\n\npixiv_query_cooldown=0  # 每次查询的冷却时间\npixiv_no_query_cooldown_users=[]  # 在这个列表中的用户不受冷却时间的影响（JSON数组）\npixiv_max_item_per_query=10  # 每个查询最多请求的插画数量\n\npixiv_tag_translation_enabled=True  # 启用搜索关键字翻译功能（平时搜索时记录标签翻译，在查询时判断是否存在对应中日翻译）\n\npixiv_more_enabled=True  # 启用重复上一次请求（还要）功能\npixiv_query_expires_in=10*60  # 上一次请求的过期时间（单位：秒）\n\npixiv_illust_query_enabled=True  # 启用插画查询（看看图）功能\n\npixiv_ranking_query_enabled=True  # 启用榜单查询（看看榜）功能\npixiv_ranking_default_mode=day  # 默认查询的榜单，可选值：day, week, month, day_male, day_female, week_original, week_rookie, day_manga\npixiv_ranking_default_range=[1, 3]  # 默认查询的榜单范围\npixiv_ranking_fetch_item=150  # 每次从服务器获取的榜单项数（查询的榜单范围必须在这个数目内）\npixiv_ranking_max_item_per_query=5  # 每次榜单查询最多能查询多少项\n\npixiv_random_illust_query_enabled=True  # 启用关键字插画随机抽选（来张xx图）功能\npixiv_random_illust_method=bookmark_proportion  # 随机抽选方法，下同，可选值：bookmark_proportion(概率与书签数成正比), view_proportion(概率与阅读量成正比), timedelta_proportion(概率与投稿时间和现在的时间差成正比), uniform(相等概率)\npixiv_random_illust_min_bookmark=0  # 过滤掉书签数小于该值的插画，下同\npixiv_random_illust_min_view=0  # 过滤掉阅读量小于该值的插画，下同\npixiv_random_illust_max_page=20  # 每次从服务器获取的查询结果页数，下同\npixiv_random_illust_max_item=500  # 每次从服务器获取的查询结果项数，下同\n\npixiv_random_recommended_illust_query_enabled=True  # 启用推荐插画随机抽选（来张图）功能\npixiv_random_recommended_illust_method=uniform\npixiv_random_recommended_illust_min_bookmark=0\npixiv_random_recommended_illust_min_view=0\npixiv_random_recommended_illust_max_page=40\npixiv_random_recommended_illust_max_item=1000\n\npixiv_random_related_illust_query_enabled=True  # 启用关联插画随机抽选（不够色）功能\npixiv_random_related_illust_method=bookmark_proportion\npixiv_random_related_illust_min_bookmark=0\npixiv_random_related_illust_min_view=0\npixiv_random_related_illust_max_page=4\npixiv_random_related_illust_max_item=100\n\npixiv_random_user_illust_query_enabled=True  # 启用用户插画随机抽选（来张xx老师的图）功能\npixiv_random_user_illust_method=timedelta_proportion\npixiv_random_user_illust_min_bookmark=0\npixiv_random_user_illust_min_view=0\npixiv_random_user_illust_max_page=2147483647\npixiv_random_user_illust_max_item=2147483647\n\npixiv_random_bookmark_query_enabled=True  # 启用用户书签随机抽选（来张私家车）功能\npixiv_random_bookmark_user_id=0  # 当QQ用户未绑定Pixiv账号时，从该Pixiv账号的书签内抽选\npixiv_random_bookmark_method=uniform\npixiv_random_bookmark_min_bookmark=0\npixiv_random_bookmark_min_view=0\npixiv_random_bookmark_max_page=2147483647\npixiv_random_bookmark_max_item=2147483647\n\npixiv_watch_interval=7200  # 更新推送的查询间隔\n```\n\n## Special Thanks\n\n[Mikubill/pixivpy-async](https://github.com/Mikubill/pixivpy-async)\n\n[nonebot/nonebot2](https://github.com/nonebot/nonebot2)\n\n## LICENSE\n\n```\nMIT License\n\nCopyright (c) 2021 ssttkkl\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the "Software"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all\ncopies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE\nSOFTWARE.\n\n```\n',
    'author': 'ssttkkl',
    'author_email': 'huang.wen.long@hotmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ssttkkl/nonebot-plugin-pixivbot',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Any

from motor.motor_asyncio import AsyncIOMotorDatabase
from pymongo import ReturnDocument

from nonebot_plugin_pixivbot import context
from .mongo_migration import MongoMigration
from .mongo_migration_manager import MongoMigrationManager


@context.require(MongoMigrationManager).register
class MongoV4ToV5(MongoMigration):
    from_db_version = 4
    to_db_version = 5

    async def inc_and_get(self, db: AsyncIOMotorDatabase, key: Any) -> int:
        seq = await db.seq.find_one_and_update({'key': key},
                                               {'$inc': {'value': 1}},
                                               upsert=True,
                                               return_document=ReturnDocument.AFTER)
        return seq["value"]

    async def migrate(self, db: AsyncIOMotorDatabase):
        await self.migrate_subscription(db)
        await self.migrate_watch_task(db)

    async def migrate_subscription(self, db: AsyncIOMotorDatabase):
        try:
            await db["subscription"].drop_index("subscriber_1_type_1")
        except:
            pass

        async for sub in db["subscription"].find():
            code = await self.inc_and_get(db, sub["subscriber"] | {"type": "subscription"})
            await db["subscription"].update_one(sub, {"$set": {"code": code}})

    async def migrate_watch_task(self, db: AsyncIOMotorDatabase):
        try:
            await db["watch_task"].drop_index("subscriber_1_type_1_kwargs_1")
        except:
            pass

        async for sub in db["watch_task"].find():
            code = await self.inc_and_get(db, sub["subscriber"] | {"type": "watch_task"})
            await db["watch_task"].update_one(sub, {"$set": {"code": code}})



from typing import Iterable, Optional, List, Collection

from nonebot import logger
from sqlalchemy import Column, String, select
from sqlalchemy.dialects.sqlite import insert

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.source.sql import SqlDataSource
from nonebot_plugin_pixivbot.model import Tag, Illust


@context.require(SqlDataSource).registry.mapped
class LocalTag:
    __tablename__ = "local_tag"

    name = Column(String, primary_key=True, nullable=False)
    translated_name = Column(String, nullable=False, index=True)


@context.inject
@context.register_singleton()
class SqlLocalTagRepo:
    data_source: SqlDataSource = Inject(SqlDataSource)

    async def find_by_name(self, name: str) -> Optional[Tag]:
        session = self.data_source.session()
        stmt = select(LocalTag).where(LocalTag.name == name).limit(1)
        local_tag = (await session.execute(stmt)).scalar_one_or_none()
        if local_tag is not None:
            return Tag.from_orm(local_tag)
        else:
            return None

    async def find_by_translated_name(self, translated_name: str) -> Optional[Tag]:
        session = self.data_source.session()
        stmt = select(LocalTag).where(LocalTag.translated_name == translated_name).limit(1)
        local_tag = (await session.execute(stmt)).scalar_one_or_none()
        if local_tag is not None:
            return Tag.from_orm(local_tag)
        else:
            return None

    async def update_one(self, tag: Tag):
        await self.update_many([tag])

    async def update_many(self, tags: Collection[Tag]):
        if len(tags) == 0:
            return

        session = self.data_source.session()
        stmt = insert(LocalTag).values([t.dict() for t in tags])
        stmt = stmt.on_conflict_do_update(index_elements=[LocalTag.name],
                                          set_={
                                              LocalTag.translated_name: stmt.excluded.translated_name
                                          })

        await session.execute(stmt)
        await session.commit()
        logger.info(f"[local_tag_repo] added {tags} local tags")

    async def update_from_illusts(self, illusts: Collection[Illust]):
        tags = {}
        for x in illusts:
            for t in x.tags:
                if t.translated_name:
                    tags[t.name] = t

        await self.update_many(tags.values())


__all__ = ("SqlLocalTagRepo",)



import asyncio
import json
from datetime import datetime, date

from nonebot import get_driver, logger
from sqlalchemy.ext.asyncio import create_async_engine, async_scoped_session, AsyncSession, AsyncEngine
from sqlalchemy.orm import registry, sessionmaker

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.errors import DataSourceNotReadyError
from nonebot_plugin_pixivbot.enums import DataSourceType


def default_dumps(obj):
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()

    return None


def json_serializer(obj):
    return json.dumps(obj, default=default_dumps)


@context.inject
@context.register_singleton()
class SqlDataSource:
    conf: Config = Inject(Config)

    def __init__(self):
        if self.conf.pixiv_data_source != DataSourceType.sqlite:
            raise RuntimeError("pixiv_data_source != 'sqlite'")

        self._engine = None
        self._session = None

        self._registry = registry()

        driver = get_driver()
        driver.on_startup(self.initialize)
        driver.on_shutdown(self.finalize)

    async def initialize(self):
        driver = get_driver()
        self._engine = create_async_engine(self.conf.pixiv_sql_conn_url,
                                           # 仅当debug模式时回显sql语句
                                           echo=driver.config.log_level.lower() == 'debug',
                                           future=True,
                                           json_serializer=json_serializer)

        async with self._engine.begin() as conn:
            await conn.run_sync(self._registry.metadata.create_all)

        # expire_on_commit=False will prevent attributes from being expired
        # after commit.
        session_factory = sessionmaker(
            self._engine, expire_on_commit=False, class_=AsyncSession
        )
        self._session = async_scoped_session(
            session_factory, scopefunc=asyncio.current_task)
        logger.success("data source initialized")

    async def finalize(self):
        await self._engine.dispose()

        self._engine = None
        self._session = None

        logger.success("data source disposed")

    @property
    def engine(self) -> AsyncEngine:
        if self._engine is None:
            raise DataSourceNotReadyError()
        return self._engine

    @property
    def registry(self) -> registry:
        return self._registry

    @property
    def session(self) -> async_scoped_session:
        if self._session is None:
            raise DataSourceNotReadyError()
        return self._session



from datetime import datetime, timezone
from math import ceil
from typing import TypeVar

from nonebot import logger

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.model import UserIdentifier
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from .permission_interceptor import PermissionInterceptor
from ..pkg_context import context

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.inject
@context.register_singleton()
class CooldownInterceptor(PermissionInterceptor):
    conf = Inject(Config)

    def __init__(self):
        super().__init__()
        self.last_query_time = dict[UserIdentifier[UID], datetime]()

    def has_permission(self, post_dest: PostDestination[UID, GID]) -> bool:
        if self.conf.pixiv_query_cooldown == 0:
            return True

        if not post_dest.user_id:
            logger.debug("cooldown intercept was skipped for group post")
            return True

        identifier = UserIdentifier(post_dest.adapter, post_dest.user_id)

        if str(post_dest.user_id) in self.conf.pixiv_no_query_cooldown_users \
                or str(identifier) in self.conf.pixiv_no_query_cooldown_users:
            return True

        now = datetime.now(timezone.utc)
        if identifier not in self.last_query_time:
            self.last_query_time[identifier] = now
            return True
        else:
            logger.debug(f"last query time ({identifier}): {self.last_query_time[identifier]}")
            delta = now - self.last_query_time[identifier]
            cooldown = self.conf.pixiv_query_cooldown - delta.total_seconds()
            if cooldown > 0:
                logger.debug(f"cooldown ({identifier}): {cooldown}s")
                return False
            else:
                self.last_query_time[identifier] = now
                return True

    def get_permission_denied_msg(self, post_dest: PostDestination[UID, GID]) -> str:
        identifier = UserIdentifier(post_dest.adapter, post_dest.user_id)
        now = datetime.now(timezone.utc)
        delta = now - self.last_query_time[identifier]
        cooldown = ceil(self.conf.pixiv_query_cooldown - delta.total_seconds())
        return f"你的CD还有{cooldown}s转好"


__all__ = ("CooldownInterceptor",)



from typing import TypeVar, Optional

from sqlalchemy import Column, String, Integer, select, delete
from sqlalchemy.dialects.sqlite import insert

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.source.sql import SqlDataSource
from nonebot_plugin_pixivbot.model import PixivBinding

UID = TypeVar("UID")


@context.require(SqlDataSource).registry.mapped
class PixivBindingOrm:
    __tablename__ = "pixiv_binding"

    adapter = Column(String, nullable=False, primary_key=True)
    user_id = Column(String, nullable=False, primary_key=True)
    pixiv_user_id = Column(Integer, nullable=False)


@context.inject
@context.register_singleton()
class SqlPixivBindingRepo:
    data_source: SqlDataSource = Inject(SqlDataSource)

    async def get(self, adapter: str, user_id: UID) -> Optional[PixivBinding]:
        session = self.data_source.session()
        stmt = (select(PixivBindingOrm)
                .where(PixivBindingOrm.adapter == adapter,
                       PixivBindingOrm.user_id == str(user_id))
                .limit(1))
        result = (await session.execute(stmt)).scalar_one_or_none()
        if result is not None:
            return PixivBinding(adapter=result.adapter,
                                user_id=type(user_id)(result.user_id),
                                pixiv_user_id=result.pixiv_user_id)
        else:
            return None

    async def update(self, binding: PixivBinding):
        session = self.data_source.session()
        stmt = (insert(PixivBindingOrm)
                .values(**binding.dict())
                .on_conflict_do_update(index_elements=[PixivBindingOrm.adapter, PixivBindingOrm.user_id],
                                       set_={
                                           PixivBindingOrm.pixiv_user_id: binding.pixiv_user_id
                                       }))
        await session.execute(stmt)
        await session.commit()

    async def remove(self, adapter: str, user_id: UID) -> bool:
        session = self.data_source.session()
        stmt = (delete(PixivBinding)
                .where(PixivBindingOrm.adapter == adapter,
                       PixivBindingOrm.user_id == str(user_id)))
        result = await session.execute(stmt)
        await session.commit()
        return result.rowcount == 1


__all__ = ("SqlPixivBindingRepo",)



from enum import Enum

__all__ = ("DataSourceType", "BlockAction", "DownloadQuantity", "RandomIllustMethod", "RankingMode",)


class DataSourceType(str, Enum):
    mongo = 'mongo'
    sqlite = 'sqlite'


class BlockAction(str, Enum):
    no_image = 'no_image'
    completely_block = 'completely_block'
    no_reply = 'no_reply'


class DownloadQuantity(str, Enum):
    original = 'original'
    square_medium = 'square_medium'
    medium = 'medium'
    large = 'large'


class RandomIllustMethod(str, Enum):
    uniform = 'uniform'
    bookmark_proportion = 'bookmark_proportion'
    view_proportion = 'view_proportion'
    timedelta_proportion = 'timedelta_proportion'


class RankingMode(str, Enum):
    day = 'day'
    week = 'week'
    month = 'month'
    day_male = 'day_male'
    day_female = 'day_female'
    week_original = 'week_original'
    week_rookie = 'week_rookie'
    day_manga = 'day_manga'



from typing import Optional, Collection

from beanie import Document
from beanie.odm.operators.update.general import SetOnInsert
from nonebot import logger
from pymongo import *

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.model import Tag, Illust
from ..source import MongoDataSource


class LocalTag(Tag, Document):
    class Settings:
        name = "local_tags"
        indexes = [
            IndexModel([("name", 1)], unique=True),
            IndexModel([("translated_name", 1)])
        ]


context.require(MongoDataSource).document_models.append(LocalTag)


@context.inject
@context.register_singleton()
class MongoLocalTagRepo:
    mongo = Inject(MongoDataSource)

    async def find_by_name(self, name: str) -> Optional[Tag]:
        result = await LocalTag.find_one(LocalTag.name == name)
        return result

    async def find_by_translated_name(self, translated_name: str) -> Optional[Tag]:
        result = await LocalTag.find_one(LocalTag.translated_name == translated_name)
        return result

    async def update_one(self, tag: Tag):
        await LocalTag.find_one(LocalTag.name == tag.name).upsert(
            SetOnInsert({LocalTag.translated_name: tag.translated_name}),
            on_insert=LocalTag(**tag.dict()),
        )

    async def update_many(self, tags: Collection[Tag]):
        # BulkWriter存在bug，upsert不生效
        # https://github.com/roman-right/beanie/issues/224
        #
        # async with BulkWriter() as bw:
        #     for tag in tags:
        #         await LocalTag.find_one(LocalTag.name == tag.name).upsert(
        #             SetOnInsert({LocalTag.translated_name: tag.translated_name}),
        #             on_insert=LocalTag(**tag.dict()),
        #             bulk_writer=bw
        #         )

        opt = []

        for tag in tags:
            opt.append(UpdateOne(
                {"name": tag.name},
                {"$setOnInsert": {"translated_name": tag.translated_name}},
                upsert=True
            ))

        if len(opt) != 0:
            await self.mongo.db.local_tags.bulk_write(opt, ordered=False)
            logger.info(f"[local_tag_repo] added {tags} local tags")

    async def update_from_illusts(self, illusts: Collection[Illust]):
        tags = {}
        for x in illusts:
            for t in x.tags:
                if t.translated_name:
                    tags[t.name] = t

        await self.update_many(tags.values())


__all__ = ("MongoLocalTagRepo",)



from typing import Optional, Protocol, Collection

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.model import Tag, Illust
from ...enums import DataSourceType


class LocalTagRepo(Protocol):
    async def find_by_name(self, name: str) -> Optional[Tag]:
        ...

    async def find_by_translated_name(self, translated_name: str) -> Optional[Tag]:
        ...

    async def update_one(self, tag: Tag):
        ...

    async def update_many(self, tags: Collection[Tag]):
        ...

    async def update_from_illusts(self, illusts: Collection[Illust]):
        ...


conf = context.require(Config)
if conf.pixiv_data_source == DataSourceType.mongo:
    from .mongo import MongoLocalTagRepo

    context.bind(LocalTagRepo, MongoLocalTagRepo)
else:
    from .sql import SqlLocalTagRepo

    context.bind(LocalTagRepo, SqlLocalTagRepo)

__all__ = ("LocalTagRepo",)



from asyncio import sleep, create_task, CancelledError, Semaphore, Task
from contextlib import asynccontextmanager
from datetime import datetime, timedelta
from functools import wraps
from io import BytesIO
from typing import TypeVar, Optional, Awaitable, List, Callable, Tuple, AsyncGenerator, Union

from nonebot import logger
from pixivpy_async import *
from pixivpy_async.error import TokenError

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.enums import DownloadQuantity, RankingMode
from nonebot_plugin_pixivbot.model import Illust, User, UserPreview
from nonebot_plugin_pixivbot.utils.errors import QueryError, RateLimitError
from nonebot_plugin_pixivbot.utils.lifecycler import on_startup, on_shutdown
from .base import PixivRepo
from .compressor import Compressor
from .lazy_illust import LazyIllust
from .models import PixivRepoMetadata

T = TypeVar("T")


def _auto_retry(func):
    @wraps(func)
    async def wrapped(*args, **kwargs):
        err = None
        for t in range(10):
            try:
                return await func(*args, **kwargs)
            except CancelledError as e:
                raise e
            except QueryError as e:
                raise e
            except Exception as e:
                logger.debug(f"Retrying... {t + 1}/10")
                logger.exception(e)
                err = e

        raise err

    return wrapped


@context.inject
@context.register_eager_singleton()
class RemotePixivRepo(PixivRepo):
    _conf = Inject(Config)
    _compressor = Inject(Compressor)

    # noinspection PyTypeChecker
    def __init__(self):
        self._sema: Semaphore = None
        self._pclient: PixivClient = None
        self._papi: AppPixivAPI = None
        self._refresh_daemon: Task = None

        self._got_rate_limit: Optional[datetime] = None

        self.user_id = 0

        on_startup(self.start, replay=True)
        on_shutdown(self.shutdown)

    async def _refresh(self):
        # Latest app version can be found using GET /old/application-info/android
        USER_AGENT = "PixivAndroidApp/5.0.234 (Android 11; Pixel 5)"
        # REDIRECT_URI = "https://app-api.pixiv.net/web/v1/users/auth/pixiv/callback"
        # LOGIN_URL = "https://app-api.pixiv.net/web/v1/login"
        AUTH_TOKEN_URL = "https://oauth.secure.pixiv.net/auth/token"
        CLIENT_ID = "MOBrBDS8blbauoSck0ZfDbtuzpyT"
        CLIENT_SECRET = "lsACyCD94FhDUtGTXi3QzcFE2uU1hqtDaKeqrdwj"

        refresh_token = self._conf.pixiv_refresh_token

        data = {
            "client_id": CLIENT_ID,
            "client_secret": CLIENT_SECRET,
            "grant_type": "refresh_token",
            "include_policy": "true",
            "refresh_token": refresh_token,
        }
        result = await self._papi.requests_(method="POST", url=AUTH_TOKEN_URL, data=data,
                                            headers={"User-Agent": USER_AGENT},
                                            auth=False)
        if result.has_error:
            raise TokenError(None, result)
        else:
            self._papi.set_auth(result.access_token, result.refresh_token)
            self.user_id = result["user"]["id"]

            logger.success(
                f"refresh access token successfully. new token expires in {result.expires_in} seconds.")
            logger.debug(f"access_token: {result.access_token}")
            logger.debug(f"refresh_token: {result.refresh_token}")

            # maybe the refresh token will be changed (even thought i haven't seen it yet)
            if result.refresh_token != refresh_token:
                refresh_token = result.refresh_token
                logger.warning(
                    f"refresh token has been changed: {result.refresh_token}")

            return result

    async def _refresh_daemon_worker(self):
        while True:
            try:
                result = await self._refresh()
                await sleep(result.expires_in * 0.8)
            except CancelledError as e:
                raise e
            except Exception as e:
                logger.error(
                    "failed to refresh access token, will retry in 60s.")
                logger.exception(e)
                await sleep(60)

    def start(self):
        self._pclient = PixivClient(proxy=self._conf.pixiv_proxy)
        self._papi = AppPixivAPI(client=self._pclient.start())
        self._papi.set_additional_headers({'Accept-Language': 'zh-CN'})
        self._refresh_daemon = create_task(self._refresh_daemon_worker())
        self._sema = Semaphore(self._conf.pixiv_simultaneous_query)

    async def shutdown(self):
        await self._pclient.close()
        self._refresh_daemon.cancel()

    def _check_error_in_raw_result(self, raw_result: dict):
        if "error" in raw_result:
            message = raw_result["error"]["user_message"] \
                      or raw_result["error"]["message"] \
                      or raw_result["error"]["reason"]
            if message == "Rate Limit":
                # 处理RateLimit
                self._got_rate_limit = datetime.utcnow()
                raise RateLimitError()
            else:
                raise QueryError(message)

    @asynccontextmanager
    async def _query(self):
        # RateLimit期间不进行查询
        if self._got_rate_limit is not None:
            now = datetime.utcnow()
            if now - self._got_rate_limit >= timedelta(minutes=2):
                self._got_rate_limit = None
            else:
                raise RateLimitError()

        async with self._sema:
            yield

    async def _load_raw_page(self, papi_search_func: Callable[..., Awaitable[dict]],
                             **kwargs):
        async with self._query():
            raw_result = await papi_search_func(**kwargs)
            self._check_error_in_raw_result(raw_result)
            return raw_result

    async def _load_page(self, papi_search_func: Callable[..., Awaitable[dict]],
                         element_list_name: str,
                         *, mapper: Optional[Callable[[dict], T]] = None,
                         filter_item: Optional[Callable[[T], bool]] = None,
                         **kwargs) -> Tuple[List[T], PixivRepoMetadata]:
        """
        加载一页
        :param papi_search_func: PixivPy-Async的加载方法
        :param element_list_name: 返回JSON中元素所在列表名
        :param mapper: 将元素从JSON格式映射为特定格式
        :param filter_item: 过滤不符合条件的元素（先映射再过滤）
        :param kwargs: 传给papi_search_func的参数
        :return: 加载结果
        """
        raw_result = await self._load_raw_page(papi_search_func, **kwargs)

        pending = []
        for x in raw_result[element_list_name]:
            if mapper is not None:
                x = mapper(x)

            if not filter_item or filter_item(x):
                pending.append(x)

        metadata = PixivRepoMetadata(next_qs=AppPixivAPI.parse_qs(next_url=raw_result["next_url"]))

        return pending, metadata

    async def _load_many_pages(self, papi_search_func: Callable[..., Awaitable[dict]],
                               element_list_name: str,
                               *, mapper: Optional[Callable[[dict], T]] = None,
                               filter_item: Optional[Callable[[T], bool]] = None,
                               **kwargs) -> AsyncGenerator[Tuple[List[T], PixivRepoMetadata], None]:
        """
        一次加载多页
        :param papi_search_func: PixivPy-Async的加载方法
        :param element_list_name: 返回JSON中元素所在列表名
        :param mapper: 将元素从JSON格式映射为特定格式
        :param filter_item: 过滤不符合条件的元素（先映射再过滤）
        :param kwargs: 传给papi_search_func的参数
        :return: 加载结果
        """
        loaded_items = 0
        loaded_pages = 0
        next_qs = kwargs

        while True:
            page, metadata = await self._load_page(papi_search_func, element_list_name, mapper=mapper,
                                                   filter_item=filter_item, **next_qs)

            loaded_pages = loaded_pages + 1
            loaded_items += len(page)
            logger.debug(f"[remote] {loaded_pages} pages loaded ({loaded_items} items in total)")

            metadata.pages = loaded_pages
            yield page, metadata

            next_qs = metadata.next_qs
            if next_qs:
                if 'viewed' in next_qs:
                    # 由于pixivpy-async的illust_recommended的bug，需要删掉这个参数
                    del next_qs['viewed']
            else:
                break

    async def _get_illusts(self, papi_search_func: Callable[[], Awaitable[dict]],
                           *, block_tags: Optional[List[str]] = None,
                           min_bookmark: int = 0,
                           min_view: int = 0,
                           **kwargs) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        """
        加载插画
        :param papi_search_func: PixivPy-Async的加载方法
        :param block_tags: 要过滤的标签
        :param min_bookmark: 书签数下限
        :param min_view: 阅读数下限
        :param kwargs: 传给papi_search_func的参数
        :return:
        """

        def illust_filter(illust: Illust) -> bool:
            # 标签过滤
            if block_tags is not None:
                for tag in block_tags:
                    if illust.has_tag(tag):
                        return False
            # 书签下限过滤
            if illust.total_bookmarks < min_bookmark:
                return False
            # 浏览量下限过滤
            if illust.total_view < min_view:
                return False
            return True

        total = 0
        broken = 0

        yield PixivRepoMetadata(pages=0, next_qs=kwargs)
        async for page, metadata in self._load_many_pages(papi_search_func, "illusts",
                                                          mapper=lambda x: Illust.parse_obj(x),
                                                          filter_item=illust_filter, **kwargs):
            for item in page:
                total += 1
                if "limit_unknown_360.png" in item.image_urls.large:
                    broken += 1
                    yield LazyIllust(item.id)
                else:
                    yield LazyIllust(item.id, item)
            yield metadata

    async def _get_user_previews(self, papi_search_func: Callable[[], Awaitable[dict]],
                                 *, block_tags: Optional[List[str]] = None,
                                 **kwargs) \
            -> AsyncGenerator[Union[PixivRepoMetadata, UserPreview], None]:
        yield PixivRepoMetadata(pages=0, next_qs=kwargs)
        async for page, metadata in self._load_many_pages(papi_search_func, "user_previews",
                                                          mapper=lambda x: UserPreview.parse_obj(x), **kwargs):
            for item in page:
                item: UserPreview
                if block_tags is not None:
                    illusts = []
                    for x in item.illusts:
                        for tag in block_tags:
                            if x.has_tag(tag):
                                break
                        else:
                            illusts.append(x)
                    item.illusts = illusts
                yield item
            yield metadata

    async def _get_users(self, papi_search_func: Callable[[], Awaitable[dict]], **kwargs) \
            -> AsyncGenerator[Union[PixivRepoMetadata, User], None]:
        yield PixivRepoMetadata(pages=0, next_qs=kwargs)
        async for page, metadata in self._load_many_pages(papi_search_func, "user_previews",
                                                          mapper=lambda x: User.parse_obj(x["user"]), **kwargs):
            for item in page:
                yield item
            yield metadata

    @_auto_retry
    async def _raw_illust_detail(self, illust_id: int, **kwargs) -> dict:
        async with self._query():
            raw_result = await self._papi.illust_detail(illust_id, **kwargs)
            self._check_error_in_raw_result(raw_result)
            return raw_result

    async def illust_detail(self, illust_id: int, **kwargs) -> AsyncGenerator[Illust, None]:
        logger.debug(f"[remote] illust_detail {illust_id}")
        raw_result = await self._raw_illust_detail(illust_id, **kwargs)
        yield PixivRepoMetadata()
        yield Illust.parse_obj(raw_result["illust"])

    @_auto_retry
    async def _raw_user_detail(self, user_id: int, **kwargs) -> dict:
        async with self._query():
            raw_result = await self._papi.user_detail(user_id, **kwargs)
            self._check_error_in_raw_result(raw_result)
            return raw_result

    async def user_detail(self, user_id: int, **kwargs) -> AsyncGenerator[User, None]:
        logger.debug(f"[remote] user_detail {user_id}")
        raw_result = await self._raw_user_detail(user_id, **kwargs)
        yield PixivRepoMetadata()
        yield User.parse_obj(raw_result["user"])

    def search_illust(self, word: str, **kwargs) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[remote] search_illust {word}")
        return self._get_illusts(self._papi.search_illust,
                                 block_tags=self._conf.pixiv_block_tags,
                                 min_bookmark=self._conf.pixiv_random_illust_min_bookmark,
                                 min_view=self._conf.pixiv_random_illust_min_view,
                                 word=word, **kwargs)

    def search_user(self, word: str, **kwargs) \
            -> AsyncGenerator[Union[User, PixivRepoMetadata], None]:
        logger.debug(f"[remote] search_user {word}")
        return self._get_users(self._papi.search_user,
                               word=word, **kwargs)

    def search_user_with_preview(self, word: str, **kwargs) \
            -> AsyncGenerator[Union[UserPreview, PixivRepoMetadata], None]:
        logger.debug(f"[remote] search_user {word}")
        return self._get_user_previews(self._papi.search_user,
                                       block_tags=self._conf.pixiv_block_tags,
                                       word=word, **kwargs)

    def user_following(self, user_id: int, **kwargs) \
            -> AsyncGenerator[Union[User, PixivRepoMetadata], None]:
        logger.debug(f"[remote] following_users {user_id}")
        return self._get_users(self._papi.user_following,
                               user_id=user_id, **kwargs)

    def user_following_with_preview(self, user_id: int, **kwargs) \
            -> AsyncGenerator[Union[UserPreview, PixivRepoMetadata], None]:
        logger.debug(f"[remote] following_users {user_id}")
        return self._get_user_previews(self._papi.user_following,
                                       block_tags=self._conf.pixiv_block_tags,
                                       user_id=user_id, **kwargs)

    def user_illusts(self, user_id: int, **kwargs) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[remote] user_illusts {user_id}")
        return self._get_illusts(self._papi.user_illusts,
                                 block_tags=self._conf.pixiv_block_tags,
                                 min_bookmark=self._conf.pixiv_random_user_illust_min_bookmark,
                                 min_view=self._conf.pixiv_random_user_illust_min_view,
                                 user_id=user_id, **kwargs)

    def user_bookmarks(self, user_id: int = 0, **kwargs) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        if user_id == 0:
            user_id = self.user_id

        logger.debug(f"[remote] user_bookmarks {user_id}")
        return self._get_illusts(self._papi.user_bookmarks_illust,
                                 block_tags=self._conf.pixiv_block_tags,
                                 min_bookmark=self._conf.pixiv_random_bookmark_min_bookmark,
                                 min_view=self._conf.pixiv_random_bookmark_min_view,
                                 user_id=user_id, **kwargs)

    # def following_illusts(self, **kwargs) \
    #         -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
    #     logger.debug(f"[remote] following_illusts")
    #     return self._get_illusts(self._papi.illust_follow,
    #                              block_tags=self._conf.pixiv_block_tags,
    #                              min_bookmark=self._conf.pixiv_random_following_illust_min_bookmark,
    #                              min_view=self._conf.pixiv_random_following_illust_min_view,
    #                              **kwargs)

    def recommended_illusts(self, **kwargs) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[remote] recommended_illusts")
        return self._get_illusts(self._papi.illust_recommended,
                                 block_tags=self._conf.pixiv_block_tags,
                                 min_bookmark=self._conf.pixiv_random_recommended_illust_min_bookmark,
                                 min_view=self._conf.pixiv_random_recommended_illust_min_view,
                                 **kwargs)

    def related_illusts(self, illust_id: int, **kwargs) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        logger.debug(f"[remote] related_illusts {illust_id}")
        return self._get_illusts(self._papi.illust_related,
                                 block_tags=self._conf.pixiv_block_tags,
                                 min_bookmark=self._conf.pixiv_random_related_illust_min_bookmark,
                                 min_view=self._conf.pixiv_random_related_illust_min_view,
                                 illust_id=illust_id, **kwargs)

    def illust_ranking(self, mode: Union[str, RankingMode], **kwargs) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        if isinstance(mode, str):
            mode = RankingMode[mode]

        logger.debug(f"[remote] illust_ranking {mode}")
        return self._get_illusts(self._papi.illust_ranking,
                                 block_tags=self._conf.pixiv_block_tags,
                                 mode=mode.name, **kwargs)

    @_auto_retry
    async def _raw_image(self, illust: Illust, **kwargs) -> bytes:
        download_quantity = self._conf.pixiv_download_quantity
        custom_domain = self._conf.pixiv_download_custom_domain

        if download_quantity == DownloadQuantity.original:
            if len(illust.meta_pages) > 0:
                url = illust.meta_pages[0].image_urls.original
            else:
                url = illust.meta_single_page.original_image_url
        else:
            url = getattr(illust.image_urls, download_quantity.name)

        if custom_domain is not None:
            url = url.replace("i.pximg.net", custom_domain)

        async with self._query():
            with BytesIO() as bio:
                await self._papi.download(url, fname=bio, **kwargs)
                content = bio.getvalue()
                return content

    async def image(self, illust: Illust, **kwargs) \
            -> AsyncGenerator[Union[bytes, PixivRepoMetadata], None]:
        logger.debug(f"[remote] image {illust.id}")
        content = await self._raw_image(illust, **kwargs)
        content = await self._compressor.compress(content)
        yield PixivRepoMetadata()
        yield content


__all__ = ("RemotePixivRepo",)
